import datetime
import inspect
import logging
import sys
import os
import getpass
import psutil
from importlib import util as importlib_util
import time
import re
from pathlib import Path
import timeit

global logger
global logging_is_on_flag

global logger_datetime
global logger_iso_format_datetime
global logger_database_format_datetime
global my_logger_program_name_for_database_routine

global logger_to_file
global logging_level
global qetl_code_dir        # Parent of qualys_etl directory
global qetl_code_dir_etld_cred_yaml_template_path
global qetl_code_dir_child  # qualys_etl directory
global qetl_pip_installed_version  # qetl installed through pip
import qualys_etl

logging_is_on_flag = False


def set_qetl_code_dir_etld_cred_yaml_template_path():
    global qetl_code_dir_etld_cred_yaml_template_path
    qetl_code_dir_etld_cred_yaml_template_path = \
        Path(qetl_code_dir, "qualys_etl", "etld_templates", ".etld_cred.yaml")

def logger_start_and_stop_function_name_decorator(f):
    #
    # Experimental sending start and finish to logger.
    # @logger_start_and_stop_function_name_decorator
    # def fun_name():
    #     print("blah")
    #
    def new_f(*args, **kwargs):
        logger.info("starting", f.__name__)
        f(*args, **kwargs)
        logger.info("finished", f.__name__)
    return new_f

def check_python_version():
    py_version = sys.version.split('\n')
    try:
        if (sys.version_info[0] >= 3) and (sys.version_info[1] >= 8):
            logger.info(f"Python version found is: {py_version}")
        else:
            logger.info("Error: sys.version.info failed.  Please use Python version 3.8 or greater.")
            raise ValueError(f"Python version < 3.8 found: {py_version}")
    except Exception as e:
        logger.error(f"Please install a version of python that can work with this product.")
        logger.error(f"Exception: {e}")
        exit(1)


def get_file_size(path_to_file):
    if Path(path_to_file).is_file():
        return Path(path_to_file).stat().st_size


def get_file_mtime(path_to_file):
    if Path(path_to_file).is_file():
        statinfo = Path(path_to_file).stat()
        return statinfo.st_mtime


def get_sqlite_version():
    global logger
    import sqlite3
    version_info = sqlite3.sqlite_version_info
    if (version_info[0] >= 3) and (version_info[1] >= 26):
        logger.info(f"SQLite version found is: {sqlite3.sqlite_version}.")
    else:
        logger.error(f"SQLite version {sqlite3.sqlite_version} is older than 3.31. Please upgrade sqlite.")
        exit(1)

    return sqlite3.version


def setup_logging_stdout(log_level=logging.INFO, my_logger_prog_name=None):
    global logger
    global logging_is_on_flag
    global logger_datetime
    global logger_iso_format_datetime
    global logger_database_format_datetime
    global my_logger_program_name_for_database_routine

    logging_is_on_flag = True
    logging.Formatter.converter = time.gmtime
    d = datetime.datetime.utcnow()
    td = f"{d.year}{d.month:02d}{d.day:02d}{d.hour:02d}{d.minute:02d}{d.second:02d}"
    logger_datetime = td
    logger_iso_format_datetime = f"{d.year}-{d.month:02d}-{d.day:02d}T{d.hour:02d}:{d.minute:02d}:{d.second:02d}Z"
    logger_database_format_datetime = f"{d.year}-{d.month:02d}-{d.day:02d} {d.hour:02d}:{d.minute:02d}:{d.second:02d}"
    my_logger_program_name_for_database_routine = my_logger_prog_name
    username = getpass.getuser()
    prog = Path(__file__).name
    if my_logger_prog_name is not None:
        prog = my_logger_prog_name

    prog = f"{prog}: {td}"

    logging.basicConfig(format=f"%(asctime)s | %(levelname)-8s | {prog:54s} | {username:15} | %(funcName)-60s | %(message)s",
                        level=log_level,)

    logger = logging.getLogger()  # Useful in qetl_manage_user when we want to set the name.
    logger.info(f"PROGRAM:     {sys.argv}")
    logger.info(f"QUALYSETL VERSION: {qualys_etl.__version__}")
    logger.info(f"LOGGING SUCCESSFULLY SETUP FOR STREAMING")


def lineno():
    return inspect.currentframe().f_back.f_lineno


def check_modules():
    try:
        import requests
        # 2023-05-13 import oschmod
        import yaml
        import xmltodict
        # 2023-05-18 import boto3
        import base64
        import shutil
        import chardet
    except ImportError as e:
        logger.error(f"Missing Required Module: {e}")
        logger.error(f"Please review installation instructions and ensure you have all required modules installed.")
        exit(1)


def set_qetl_code_dir(log=True): # Module Directories
    global qetl_code_dir         # Parent of qualys_etl directory
    global qetl_code_dir_child   # qualys_etl directory

    test_exec_for_qetl_code_dir = __file__
    test_spec_for_qetl_code_dir = importlib_util.find_spec("qualys_etl")  # Installed on system

    result = ""
    if test_exec_for_qetl_code_dir.__contains__("qualys_etl"):
        result = re.sub("qualys_etl.*", '', test_exec_for_qetl_code_dir)
    elif test_spec_for_qetl_code_dir is not None:
        result = re.sub("qualys_etl.*", '', test_spec_for_qetl_code_dir.origin)
    else:
        logger.error(f"test_exec_for_qetl_code_dir - {test_exec_for_qetl_code_dir}")
        logger.error(f"test_spec_for_qetl_code_dir  - {test_spec_for_qetl_code_dir}")
        logger.error(f"Could not determine qetl code directory location.")
        logger.error(f"Please execute qetl_manage_users.py to test user")
        exit(1)

    # Module Directories
    qetl_code_dir = Path(result)
    qetl_code_dir_child = Path(qetl_code_dir, "qualys_etl")
    qetl_code_dir_child_api_host_list = Path(qetl_code_dir_child, "etld_host_list")
    qetl_code_dir_child_api_knowledgebase = Path(qetl_code_dir_child, "etld_knowledgebase")
    qetl_code_dir_child_api_lib = Path(qetl_code_dir_child, "etld_lib")
    qetl_code_dir_child_api_templates = Path(qetl_code_dir_child, "etld_templates")

    # Ensure modules are on sys.path
    modules = [qetl_code_dir_child, qetl_code_dir_child_api_lib, qetl_code_dir_child_api_templates,
               qetl_code_dir_child_api_knowledgebase, qetl_code_dir_child_api_host_list]
    for path in modules:
        if not sys.path.__contains__(str(path.absolute())):
            sys.path.insert(0, str(path))

    logger.info(f"qualysetl app dir    - {qetl_code_dir}")
    logger.info(f"qualys_etl code dir  - {qetl_code_dir_child}")
    logger.info(f"etld_lib             - {qetl_code_dir_child_api_lib}")
    logger.info(f"etld_templates       - {qetl_code_dir_child_api_templates}")
    logger.info(f"etld_knowledgebase   - {qetl_code_dir_child_api_knowledgebase}")
    logger.info(f"etld_host_list        - {qetl_code_dir_child_api_host_list}")


def get_formatted_file_info_dict(file_name):
    file_path = Path(file_name)
    if file_path.is_file():
        file_size = human_readable_size(Path(file_name).stat().st_size)
        file_change_time = Path(file_name).stat().st_ctime
        d = datetime.datetime.fromtimestamp(file_change_time)
        td = f"{d.year}-{d.month:02d}-{d.day:02d} {d.hour:02d}:{d.minute:02d}:{d.second:02d} local timezone"
        return {'file_size': file_size, 'file_change_time': td}
    else:
        return {'file_size': '', 'file_change_time': ''}


def human_readable_size(size_in_bytes):
    my_bytes = float(size_in_bytes)
    kilobytes = float(1024)
    megabytes = float(kilobytes ** 2)
    gigabytes = float(kilobytes ** 3)
    terabytes = float(kilobytes ** 4)
    petabytes = float(kilobytes ** 5)

    if my_bytes < kilobytes:
        message = 'bytes' if 0 == my_bytes > 1 else 'byte'
        return f'{my_bytes} {message}'
    elif kilobytes <= my_bytes < megabytes:
        return f'{(my_bytes / kilobytes):0.2f} kilobytes'
    elif megabytes <= my_bytes < gigabytes:
        return f'{(my_bytes / megabytes):0.2f} megabytes'
    elif gigabytes <= my_bytes < terabytes:
        return f'{(my_bytes / gigabytes):0.2f} gigabytes'
    elif terabytes <= my_bytes:
        return f'{(my_bytes / terabytes):0.2f} terabytes'
    elif petabytes <= my_bytes:
        return f'{(my_bytes / petabytes):0.2f} petabytes'


def log_file_info(file_name, msg1='output file'):
    file_info = get_formatted_file_info_dict(file_name)
    logger.info(f"{msg1} - {str(file_name)} size: {file_info.get('file_size')} "
                f"change time: {file_info.get('file_change_time')}")




class DisplayCounterToLog:
    # TODO Create method to display counters as json.
    # {"Q_Host_List": "283,129", "Q_Host_List_Detection_HOSTS": "171,452","Q_Host_List_Detection_QIDS": "1,335,871"}
    def __init__(self, display_counter_at=10000, logger_func=None, display_counter_log_message="count"):
        self.counter = 0
        self.display_counter = 0
        self.batch_name = ''
        self.batch_number = 0
        self.batch_date = ''
        self.status_table_name = 'TABLE_NAME'
        self.status_name_column = 'TABLE_NAME_LOAD_STATUS'
        self.display_counter_at = display_counter_at
        self.logger_func = logger_func
        self.display_counter_log_message = display_counter_log_message

    def update_counter(self):
        self.counter += 1
        self.display_counter += 1

    def get_counter(self):
        return self.counter

    def get_batch_name(self):
        return self.batch_name

    def set_batch_name(self, batch_name):
        self.batch_name = batch_name

    def get_batch_number(self):
        return self.batch_number

    def get_batch_date(self):
        return self.batch_date

    def set_batch_date(self, batch_date):
        self.batch_date = batch_date

    def get_status_table_name(self):
        return self.status_table_name

    def get_status_name_column(self):
        return self.status_name_column

    def update_batch_info(self, batch_name, batch_number, batch_date, status_table_name, status_name_column):
        self.batch_name = batch_name
        self.batch_number = batch_number
        self.batch_date = batch_date
        self.status_table_name = status_table_name
        self.status_name_column = status_name_column

    def get_batch_info(self):
        return {
            'batch_name': self.batch_name,
            'batch_number': self.batch_number,
            'batch_date': self.batch_date,
            'etl_workflow_status_table_name': self.status_table_name,
            'status_name_column': self.status_name_column
        }

    def update_counter_and_display_to_log(self):
        self.update_counter()
        if self.display_counter >= self.display_counter_at:
            self.display_counter = 0
            if self.logger_func is None:
                print(f"{self.display_counter_log_message}: {self.counter:,}")
            else:
                self.logger_func(f"{self.display_counter_log_message}: {self.counter:,}")

    def display_final_counter_to_log(self):
        if self.logger_func is None:
            print(f"{self.display_counter_log_message} - Total: {self.counter:,}")
        else:
            self.logger_func(f"{self.display_counter_log_message} - Total: {self.counter:,}")


def log_system_information(logger_method=print, data_directory="/opt/qetl/users"):
    # TODO Calculate final averages of cpu, mem usage.
    sys_stat_counters = {
        'cpu_usage_pct': get_cpu_usage_pct(),
        'cpu_frequency': get_cpu_frequency(),
        'cpu_count': get_cpu_count(),
        'ram_usage_pct': get_ram_usage_pct(),
        'ram_size_total_in_mb': get_ram_size_total_in_mb(),
        'ram_size_usage_in_mb': get_ram_size_usage_in_mb(),
        'ram_size_available_in_mb': get_ram_size_available_in_mb(),
        'swap_usage_pct': get_swap_usage_pct(),
        'swap_size_total_in_mb': get_swap_size_total_in_mb(),
        'swap_size_usage_in_mb': get_swap_size_usage_in_mb(),
        'swap_size_available_in_mb': get_swap_size_available_in_mb(),
        'disk_usage_pct': 0,
        'disk_size_total_in_mb': 0,
        'disk_size_usage_in_mb': 0,
        'disk_size_available_in_mb': 0,
    }

    logger_method(
        f'SYS STAT: CPU usage={sys_stat_counters["cpu_usage_pct"]} %, '
        f'frequency='
        f'{sys_stat_counters["cpu_frequency"]} MHz, '
        f'count='
        f'{sys_stat_counters["cpu_count"]}'
    )
    logger_method(
        f'SYS STAT: RAM usage='
        f'{sys_stat_counters["ram_usage_pct"]} %, '
        f'total='
        f'{sys_stat_counters["ram_size_total_in_mb"]:,.0f} MB, '
        f'used='
        f'{sys_stat_counters["ram_size_usage_in_mb"]:,.0f} MB, '
        f'available='
        f'{sys_stat_counters["ram_size_available_in_mb"]:,.0f} MB '
    )
    logger_method(
        f'SYS STAT: SWAP usage='
        f'{sys_stat_counters["swap_usage_pct"]} %, '
        f'total='
        f'{sys_stat_counters["swap_size_total_in_mb"]:,.0f} MB, '
        f'used='
        f'{sys_stat_counters["swap_size_usage_in_mb"]:,.0f} MB, '
        f'available='
        f'{sys_stat_counters["swap_size_available_in_mb"]:,.0f} MB '
    )

    try:
        disk_usage_pct = get_disk_usage_pct(data_dir=data_directory)
        disk_size_total_in_mb = get_disk_size_total_in_mb(data_dir=data_directory)
        disk_size_usage_in_mb = get_disk_size_usage_in_mb(data_dir=data_directory)
        disk_size_available_in_mb = get_disk_size_available_in_mb(data_dir=data_directory)
        sys_stat_counters['disk_usage_pct'] = disk_usage_pct
        sys_stat_counters['disk_size_total_in_mb'] = disk_size_total_in_mb
        sys_stat_counters['disk_size_usage_in_mb'] = disk_size_usage_in_mb
        sys_stat_counters['disk_size_available_in_mb'] = disk_size_available_in_mb
        logger_method(
            f'SYS STAT: DISK usage='
            f'{sys_stat_counters["disk_usage_pct"]} %, '
            f'total='
            f'{sys_stat_counters["disk_size_total_in_mb"]:,.0f} MB, '
            f'used='
            f'{sys_stat_counters["disk_size_usage_in_mb"]:,.0f} MB, '
            f'available='
            f'{sys_stat_counters["disk_size_available_in_mb"]:,.0f} MB, '
            f'path='
            f'{data_directory}, '
        )
    except Exception as e:
        pass

    return sys_stat_counters


def if_disk_space_usage_greater_than_90_log_warning(logger_method=print, data_directory="/opt/qetl/users"):
    try:
        disk_size_usage_pct = int(get_disk_usage_pct(data_dir=data_directory))
        if disk_size_usage_pct > 90:
            logger_method(f'SYS WARNING: DISK percent usage > 90% at {disk_size_usage_pct}%')
    except Exception as e:
        pass


def if_swap_space_total_is_zero_abort(logger_method=print):
    if int(get_swap_size_total_in_mb()) > 0:
        pass
    else:
        logger_method(f'SYS ERROR: Please contact your systems administrator to configure swap space.')
        logger_method(f'SYS ERROR: current swap space size is: {get_swap_size_total_in_mb():,.0f} MB, ')
        exit(1)


def if_swap_space_total_is_low_log_warning(logger_method=print):
    if int(get_swap_size_total_in_mb()) < 4000:
        logger_method(f'SYS WARNING: SWAP size total is < 4 Gig. '
                      f'Consider increasing SWAP to support larger batch jobs and sql queries')


def if_ram_space_size_is_low_log_warning(logger_method=print):
    ram_size_total_in_mb = int(get_ram_size_total_in_mb())
    if ram_size_total_in_mb < 8100:
        logger_method(f'SYS WARNING: RAM size total < 8 Gig.  '
                      f'Consider increasing RAM to support larger batch jobs and sql queries')


def get_disk_usage_pct(data_dir="/opt/qetl/users"):
    return psutil.disk_usage(data_dir).percent


def get_disk_size_total_in_mb(data_dir="/opt/qetl/users"):
    return int(psutil.disk_usage(data_dir).total / 1024 / 1024)


def get_disk_size_usage_in_mb(data_dir="/opt/qetl/users"):
    return int(psutil.disk_usage(data_dir).used / 1024 / 1024)


def get_disk_size_available_in_mb(data_dir="/opt/qetl/users"):
    total_disk_space = psutil.disk_usage(data_dir).total / 1024 / 1024
    used_disk_space = psutil.disk_usage(data_dir).used / 1024 / 1024
    available_disk_space = total_disk_space - used_disk_space
    return int(available_disk_space)


def get_cpu_usage_pct():
    return psutil.cpu_percent(interval=0.5)


def get_cpu_frequency():
    return int(psutil.cpu_freq().current)


def get_cpu_count():
    return int(psutil.cpu_count())


def get_ram_size_usage_in_mb():
    return int(psutil.virtual_memory().total - psutil.virtual_memory().available) / 1024 / 1024


def get_ram_size_total_in_mb():
    return int(psutil.virtual_memory().total) / 1024 / 1024


def get_ram_usage_pct():
    return psutil.virtual_memory().percent


def get_ram_size_available_in_mb():
    return int(psutil.virtual_memory().available) / 1024 / 1034


def get_swap_size_usage_in_mb():
    return int(psutil.swap_memory().used) / 1024 / 1024


def get_swap_size_total_in_mb():
    return int(psutil.swap_memory().total) / 1024 / 1024


def get_swap_size_available_in_mb():
    total_swap_space = psutil.swap_memory().total / 1024 / 1024
    used_swap_space = psutil.swap_memory().used / 1024 / 1024
    available_swap_space = total_swap_space - used_swap_space
    return int(available_swap_space)


def get_swap_usage_pct():
    return psutil.swap_memory().percent


def main(log_level=logging.INFO, my_logger_prog_name=None):
    global logging_level
    global qetl_code_dir_etld_cred_yaml_template_path
    setup_logging_stdout(log_level, my_logger_prog_name)
    check_modules()
    check_python_version()
    get_sqlite_version()
    set_qetl_code_dir()
    set_qetl_code_dir_etld_cred_yaml_template_path()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# Manage qetl users.
import argparse
import os
import sys
import select
import re
import logging
import time
import getpass
import json
from pathlib import Path
import qualys_etl
from qualys_etl.etld_lib import etld_lib_functions
from qualys_etl.etld_lib import etld_lib_config
from qualys_etl.etld_lib import etld_lib_credentials
from qualys_etl.etld_lib import etld_lib_authentication_objects
from qualys_etl.etld_lib import etld_lib_datetime

global qetl_user_home
global found_api_credentials
global command_line_arguments


def validate_username(username):
    username_match = re.fullmatch(r"[-_A-Za-z0-9]+", username)
    if username_match is None:
        return False
    else:
        return True


def validate_api_fqdn_server(api_fqdn_server):
    api_fqdn_server_match = re.fullmatch(r"^.*qualysapi\..*$", api_fqdn_server)  # Matches qualysapi.
    if api_fqdn_server_match is None:
        return False
    else:
        return True


def validate_gateway_fqdn_server(gateway_fqdn_server):
    gateway_fqdn_server_match = re.fullmatch(r"^.*gateway\..*$", gateway_fqdn_server)  # Matches gateway
    if gateway_fqdn_server_match is None:
        return False
    else:
        return True


def validate_password(password):
    password_match = re.fullmatch(r"[ ]+", password)
    if password_match is None:
        return True  # found no spaces in password, accept it for now.
    else:
        return False  # found spaces in password


def update_api_fqdn_server(cred):
    print(f"Current api_fqdn_server: {cred.get('api_fqdn_server')}")
    response = input(f"Update api_fqdn_server? ( yes or no ): ")
    if response == 'yes':
        while True:
            new_api_fqdn_server = input(f"Enter new api_fqdn_server: ")
            new_api_fqdn_server = re.sub("https://", '', new_api_fqdn_server)
            new_api_fqdn_server = re.sub("http://", '', new_api_fqdn_server)
            new_api_fqdn_server = re.sub("/", '', new_api_fqdn_server)
            if validate_api_fqdn_server(new_api_fqdn_server):
                break
            else:
                print(f"Not valid qualysapi FQDN, retry.")
        cred['api_fqdn_server'] = new_api_fqdn_server
    pass


def update_gateway_fqdn_server(cred):
    #
    if cred.get('api_fqdn_server') in etld_lib_credentials.platform_url.keys():
        cred['gateway_fqdn_server'] = etld_lib_credentials.platform_url.get(cred.get('api_fqdn_server'))

    print(f"Current gateway_fqdn_server: {cred.get('gateway_fqdn_server')}")
    response = input(f"Update gateway_fqdn_server? ( yes or no ): ")
    if response == 'yes':
        while True:
            new_gateway_fqdn_server = input(f"Enter new gateway_fqdn_server: ")
            new_gateway_fqdn_server = re.sub("https://", '', new_gateway_fqdn_server)
            new_gateway_fqdn_server = re.sub("http://", '', new_gateway_fqdn_server)
            new_gateway_fqdn_server = re.sub("/", '', new_gateway_fqdn_server)
            if validate_gateway_fqdn_server(new_gateway_fqdn_server):
                break
            else:
                print(f"Not valid gateway FQDN, retry.")
        cred['gateway_fqdn_server'] = new_gateway_fqdn_server
    pass


def update_gateway_fqdn_server_from_etld_lib_authentication_objects(cred, authentication_obj: etld_lib_authentication_objects.QualysAuthenticationObj):
    cred['gateway_fqdn_server'] = authentication_obj.get_gateway_platform_fqdn(cred['api_fqdn_server'])
    print(f"Current gateway_fqdn_server: {cred.get('gateway_fqdn_server')}")
    response = input(f"Update gateway_fqdn_server? ( yes or no ): ")
    if response == 'yes':
        while True:
            new_gateway_fqdn_server = input(f"Enter new gateway_fqdn_server: ")
            new_gateway_fqdn_server = re.sub("https://", '', new_gateway_fqdn_server)
            new_gateway_fqdn_server = re.sub("http://", '', new_gateway_fqdn_server)
            new_gateway_fqdn_server = re.sub("/", '', new_gateway_fqdn_server)
            if validate_gateway_fqdn_server(new_gateway_fqdn_server):
                break
            else:
                print(f"Not valid gateway FQDN, retry.")
        cred['gateway_fqdn_server'] = new_gateway_fqdn_server
    pass


def update_username(cred):
    time.sleep(1)
    print(f"\n\nCurrent username: {cred.get('username')} in config: {etld_lib_config.qetl_user_cred_file}")
    response = input(f"Update Qualys username? ( yes or no ): ")
    if response == 'yes':
        while True:
            new_username = input(f"Enter new Qualys username: ")
            if validate_username(new_username):
                break
            else:
                print(f"Found invalid characters.  Try again.  Use only AlphaNumeric or underscore in username.")
        cred['username'] = new_username
    pass


def update_password(cred):
    print(f"Update password for username: {cred.get('username')}")
    response = input(f"Update password? ( yes or no ): ")
    if response == 'yes':
        while True:
            new_password = getpass.getpass(f"Enter your Qualys password: ")
            if validate_password(new_password):
                break
            else:
                print(f"Found spaces in password, try again.")
        cred['password'] = new_password
    pass


def if_prompt_credentials():
    def my_help():
        print(f"You selected -p prompt, but the data provided is invalid.")
        print(f"Please check your values and re-enter.")

    if command_line_arguments.prompt_credentials is not None:
        while True:
            username = input("Enter username: ")
            password = getpass.getpass()
            api_fqdn_server = input("Enter api_fqdn_server: ")
            gateway_fqdn_server = input("Optionally enter gateway_fqdn_server: ")
            if validate_username(username) and validate_password(password) and validate_api_fqdn_server(api_fqdn_server):
                os.environ['q_username'] = username
                os.environ['q_password'] = password
                os.environ['q_api_fqdn_server'] = api_fqdn_server
                break
            else:
                my_help()
                continue

        if validate_gateway_fqdn_server(gateway_fqdn_server):
            os.environ['q_gateway_fqdn_server'] = gateway_fqdn_server
        else:
            os.environ['q_gateway_fqdn_server'] = ""
        return True
    else:
        return False


def valid_json(json_data):
    try:
        json.loads(json_data)
    except ValueError as err:
        return False
    return True


def if_user_selected_sending_credentials_via_stdin():
    def my_help():
        print(f"You selected -s stdin, but the data provided is invalid.")
        print('Example of valid json (gateway is optional): ')
        print('{"q_username":"your userid", "q_password":"your password", "q_api_fqdn_server":"api fqdn", '
              '"q_gateway_fqdn_server": "gateway api fqdn"}')
        print(f"Please rerun when ready.")

    if command_line_arguments.stdin_credentials is not None:
        input_fd_ready, output_fd_ready, exception_fd_ready = select.select([sys.stdin], [], [], 5)
        stdin_data = ""
        if input_fd_ready:
            for i in sys.stdin.readlines():
                stdin_data = f"{stdin_data}{i}"
            if valid_json(stdin_data):
                pass
            else:
                print("Invalid JSON...")
                my_help()
                exit(1)
        else:
            print("No data set through stdin...")
            my_help()
            exit(1)

        cred = json.loads(stdin_data)
        if 'q_username' in cred.keys() and 'q_password' in cred.keys() and 'q_api_fqdn_server' in cred.keys():
            pass
        else:
            print("Missing q_username, q_password or q_api_fqdn_server keys...")
            my_help()
            exit(1)

        api_fqdn_server = cred['q_api_fqdn_server']
        username = cred['q_username']
        password = cred['q_password']
        gateway_fqdn_server = ""
        if 'q_gateway_fqdn_server' in cred.keys():
            gateway_fqdn_server = cred['q_gateway_fqdn_server']

        if validate_username(username) and validate_password(password) and validate_api_fqdn_server(api_fqdn_server):
            os.environ['q_username'] = username
            os.environ['q_password'] = password
            os.environ['q_api_fqdn_server'] = api_fqdn_server
        else:
            print("Invalid q_username, q_password or q_api_fqdn_server...")
            my_help()
            exit(1)
        if validate_gateway_fqdn_server(gateway_fqdn_server):
            os.environ['q_gateway_fqdn_server'] = gateway_fqdn_server
        return True
    else:
        return False


def if_user_selected_setting_credentials_in_memory():
    def my_help():
        print(f"You selected -m memory, but there is an issue with your environment variables.")
        print(f"Please check your exported q_username, q_password and q_api_fqdn_server.")
        print(f"Please rerun when ready.")

    if command_line_arguments.memory_credentials is not None:
        if 'q_username' in os.environ.keys() and \
                'q_password' in os.environ.keys() and \
                'q_api_fqdn_server' in os.environ.keys():
            pass
        else:
            print("Missing q_username, q_password or q_api_fqdn_server from environment...")
            my_help()
            exit(1)

        api_fqdn_server = os.environ['q_api_fqdn_server']
        username = os.environ['q_username']
        password = os.environ['q_password']
        gateway_fqdn_server = ""
        if 'gateway_fqdn_server' in os.environ.keys():
            gateway_fqdn_server = os.environ['gateway_fqdn_server']

        if validate_username(username) and validate_password(password) and validate_api_fqdn_server(api_fqdn_server):
            pass
        else:
            print("Invalid q_username, q_password or q_api_fqdn_server in environment.  Please review your exports...")
            my_help()
            exit(1)
        if validate_gateway_fqdn_server(gateway_fqdn_server):
            os.environ['q_gateway_fqdn_server'] = gateway_fqdn_server
        return True
    else:
        return False


def if_user_selected_update_credentials_stored_on_disk(first_time=False):
    if command_line_arguments.credentials is not None or first_time is not False:
        etld_lib_config.main()
        etld_lib_credentials.main()
        credentials = etld_lib_credentials.get_cred()
        update_username(credentials)
        update_api_fqdn_server(credentials)
        update_gateway_fqdn_server(credentials)
        update_password(credentials)
        old_cred = etld_lib_credentials.get_cred()
        if old_cred == credentials:
            print(f"No changes to qualys username, password, api_fqdn_server or gateway_fqdn_server.")
        else:
            etld_lib_credentials.update_cred(credentials)
            new_credentials = etld_lib_credentials.get_cred()
            etld_lib_functions.logger.info(f"credentials updated.  username: {new_credentials.get('username')} "
                                           f"api_fqdn_server: {new_credentials.get('api_fqdn_server')} "
                                           f"gateway_fqdn_server: {new_credentials.get('gateway_fqdn_server')} "
                                           f"")
            print(f"You have updated your credentials.")
            print(f"  Qualys Username: {new_credentials.get('username')}")
            print(f"  Qualys api_fqdn_server: {new_credentials.get('api_fqdn_server')}\n")
        return True
    else:
        return False


def if_user_selected_update_credentials_stored_on_disk_from_etld_lib_authentication_objects(first_time=False):
    if Path(etld_lib_config.qetl_user_cred_file).is_file():
        pass
    else:
        first_time = True

    if command_line_arguments.credentials is not None or first_time is not False:
        etld_lib_config.main()
        etld_lib_authentication_objects.main_no_logger_for_qetl_manage_user(test_about_qualys_flag=False)
        authentication_obj: etld_lib_authentication_objects.QualysAuthenticationObj \
            = etld_lib_authentication_objects.qualys_authentication_obj
        credentials = authentication_obj.get_credentials()
        old_cred = authentication_obj.get_credentials()
        # etld_lib_credentials.main()
        # credentials = etld_lib_credentials.get_cred()
        update_username(credentials)
        update_api_fqdn_server(credentials)
        update_gateway_fqdn_server_from_etld_lib_authentication_objects(credentials, authentication_obj)
        update_password(credentials)
        # old_cred = etld_lib_credentials.get_cred()
        if old_cred == credentials:
            print(f"No changes to qualys username, password, api_fqdn_server or gateway_fqdn_server.")
        else:
            authentication_obj.update_yaml_file_credentials(new_cred=credentials)
            authentication_obj.set_credentials(force_from_file=True)
            new_credentials = authentication_obj.get_credentials()
            # etld_lib_credentials.update_cred(credentials)
            # new_credentials = etld_lib_credentials.get_cred()
            print(f"Credentials updated.  OLD: username: {old_cred.get('username')}, "
                  f"api_fqdn_server: {old_cred.get('api_fqdn_server')}, "
                  f"gateway_fqdn_server: {old_cred.get('gateway_fqdn_server')}, "
                  f"")
            print(f"Credentials updated.  NEW: username: {new_credentials.get('username')}, "
                                           f"api_fqdn_server: {new_credentials.get('api_fqdn_server')}, "
                                           f"gateway_fqdn_server: {new_credentials.get('gateway_fqdn_server')}, "
                                           f"")
            print(f"You have updated your credentials.")
            print(f"  Qualys Username:            {new_credentials.get('username')}")
            print(f"  Qualys api_fqdn_server:     {new_credentials.get('api_fqdn_server')}")
            print(f"  Qualys gateway_fqdn_server: {new_credentials.get('gateway_fqdn_server')}\n")
        return True
    else:
        return False


def start_etl_knowledgebase():
    import qualys_etl.etld_knowledgebase.knowledgebase_01_spawn_workflow_manager as etl_kb_spawn_from_qetl_manage_user
    print(f"Starting etl_knowledgebase.  For progress see your {etld_lib_config.kb_log_file}")
    etl_kb_spawn_from_qetl_manage_user.main()
    print(f"End      etl_knowledgebase.  For progress see your {etld_lib_config.kb_log_file}")


def start_etl_host_list():
    import qualys_etl.etld_host_list.host_list_01_spawn_workflow_manager \
        as etl_host_list_spawn_from_qetl_manage_user
    print(f"Starting etl_host_list.  For progress see: {etld_lib_config.host_list_log_file}")
    etl_host_list_spawn_from_qetl_manage_user.main()
    print(f"End      etl_host_list.  For results see:  {etld_lib_config.host_list_log_file}")


def start_etl_asset_inventory():
    import qualys_etl.etld_asset_inventory.asset_inventory_01_spawn_workflow_manager \
        as etl_asset_inventory_spawn_from_qetl_manage_user
    print(f"Starting etl_asset_inventory.  For progress see: {etld_lib_config.asset_inventory_log_file}")
    etl_asset_inventory_spawn_from_qetl_manage_user.main()
    print(f"End      etl_asset_inventory.  For progress see: {etld_lib_config.asset_inventory_log_file}")


def start_etl_host_list_detection():
    import qualys_etl.etld_host_list_detection.host_list_detection_01_spawn_workflow_manager \
       as etl_host_list_detection_spawn_from_qetl_manage_user
    print(f"Starting etl_host_list_detection.  For progress see: {etld_lib_config.host_list_detection_log_file}")
    etl_host_list_detection_spawn_from_qetl_manage_user.main()
    print(f"End      etl_host_list_detection.  For results see:  {etld_lib_config.host_list_detection_log_file}")


def start_etl_was():
    import qualys_etl.etld_was.was_01_spawn_workflow_manager \
        as etl_was_spawn_from_qetl_manage_user
    print(f"Starting etl_was.  For progress see: {etld_lib_config.was_log_file}")
    etl_was_spawn_from_qetl_manage_user.main()
    print(f"End      etl_was.  For results see:  {etld_lib_config.was_log_file}")

def start_etl_pcrs():
    import qualys_etl.etld_pcrs.pcrs_01_spawn_workflow_manager \
        as etl_pcrs_spawn_from_qetl_manage_user
    print(f"Starting etl_pcrs.  For progress see: {etld_lib_config.pcrs_log_file}")
    etl_pcrs_spawn_from_qetl_manage_user.main()
    print(f"End      etl_pcrs.  For progress see: {etld_lib_config.pcrs_log_file}")

def start_etl_test_system():
    from qualys_etl.etld_test_system import test_system_01_spawn_workflow_manager
    print(f"Starting etl_test_system.  For progress see your {etld_lib_config.test_system_log_file}")
    test_system_01_spawn_workflow_manager.main()
    print(f"End      etl_test_system.  For progress see your {etld_lib_config.test_system_log_file}")


def start_etl_validate(etl_option=""):
    from qualys_etl.etld_lib import etld_lib_log_validation
    try:
        if etl_option == 'validate_etl_all':
            etld_lib_log_validation.main_validate_all_logs_have_no_errors()
        elif etld_lib_log_validation.main_validate_log_has_no_errors(etl_option):
            pass
        else:
           exit(1)
    except FileNotFoundError as fnf:
        print(f"Log File Not Found: {fnf}")
        print(f"{etl_option} validation log not found, please rerun after executing corresponding {etl_option.replace('validate_', '')}")
        exit(1)
    except Exception as e:
        help_message(f"-e {etl_option} is not supported.")
        exit(1)


def validate_and_set_command_line_arguments_datetime_in_etld_lib_config():
    if command_line_arguments.datetime is None:
        pass
    else:
        if etld_lib_datetime.is_valid_qualys_datetime_format(command_line_arguments.datetime):
            etld_lib_config.qetl_manage_user_selected_datetime = command_line_arguments.datetime
        else:
            print(f"\nInvalid datetime: {str(command_line_arguments.datetime)}, "
                  f"please review format and retry when ready")
            print(f"Option: -d 'YYYY-MM-DDThh:mm:ssZ'")
            exit(1)

def etl_conflict_check(etl_module, current_etl_module):
    process_list = etld_lib_config.pgrep_af(f"{etld_lib_config.qetl_user_root_dir}.*{etl_module}")
    if len(process_list) > 0:
        print(f"Error: Conflicting ETL: {etl_module}, running for user: {etld_lib_config.qetl_user_root_dir}")
        print(f"Error: ETL: {current_etl_module} cannot execute concurrently with {etl_module} for user: {etld_lib_config.qetl_user_root_dir}")
        for process_info in process_list:
            print(f"Process Info: {process_info}")
        print(f"Error: Rerun when after Conflicting ETL completes: {etl_module} -> {etld_lib_config.qetl_user_root_dir}")
        exit(1)

def if_user_selected_execute_an_etl_module():

    validate_and_set_command_line_arguments_datetime_in_etld_lib_config()
    if str(command_line_arguments.execute_etl_module).startswith("validate_etl_"):
        start_etl_validate(etl_option=command_line_arguments.execute_etl_module)
    elif command_line_arguments.execute_etl_module == 'etl_knowledgebase':
        etl_conflict_check(etl_module='etl_knowledgebase', current_etl_module=str(command_line_arguments.execute_etl_module))
        etl_conflict_check(etl_module='etl_host_list_detection', current_etl_module=str(command_line_arguments.execute_etl_module))
        etl_conflict_check(etl_module='etl_was', current_etl_module=str(command_line_arguments.execute_etl_module))
        start_etl_knowledgebase()
    elif command_line_arguments.execute_etl_module == 'etl_host_list':
        etl_conflict_check(etl_module='etl_host_list', current_etl_module=str(command_line_arguments.execute_etl_module))
        etl_conflict_check(etl_module='etl_host_list_detection', current_etl_module=str(command_line_arguments.execute_etl_module))
        start_etl_host_list()
    elif command_line_arguments.execute_etl_module == 'etl_host_list_detection':
        etl_conflict_check(etl_module='etl_host_list_detection', current_etl_module=str(command_line_arguments.execute_etl_module))
        etl_conflict_check(etl_module='etl_host_list', current_etl_module=str(command_line_arguments.execute_etl_module))
        etl_conflict_check(etl_module='etl_knowledgebase', current_etl_module=str(command_line_arguments.execute_etl_module))
        start_etl_host_list_detection()
    elif command_line_arguments.execute_etl_module == 'etl_asset_inventory':
        etl_conflict_check(etl_module='etl_asset_inventory', current_etl_module=str(command_line_arguments.execute_etl_module))
        start_etl_asset_inventory()
    elif command_line_arguments.execute_etl_module == 'etl_was':
        etl_conflict_check(etl_module='etl_was', current_etl_module=str(command_line_arguments.execute_etl_module))
        etl_conflict_check(etl_module='etl_knowledgebase', current_etl_module=str(command_line_arguments.execute_etl_module))
        etl_conflict_check(etl_module='etl_host_list_detection', current_etl_module=str(command_line_arguments.execute_etl_module))
        start_etl_was()
    elif command_line_arguments.execute_etl_module == 'etl_pcrs':
        etl_conflict_check(etl_module='etl_pcrs', current_etl_module=str(command_line_arguments.execute_etl_module))
        start_etl_pcrs()
    elif command_line_arguments.execute_etl_module == 'etl_test_system':
        etl_conflict_check(etl_module='etl_test_system', current_etl_module=str(command_line_arguments.execute_etl_module))
        start_etl_test_system()
    elif command_line_arguments.execute_etl_module is None:
        pass
    else:
        #etld_lib_functions.logger.info(f"Invalid Option: {str(command_line_arguments.execute_etl_module)}, "
        #                               f"retry when ready.")
        print(f"\nInvalid Option: {str(command_line_arguments.execute_etl_module)}, retry when ready")
        help_message("")
        exit(1)


def if_user_selected_test_qualys_basic_authentication_login(first_time=False):
    if command_line_arguments.test is not None or first_time is not False:
        etld_lib_functions.main(log_level=logging.INFO, my_logger_prog_name='qetl_manage_user')
        etld_lib_config.main()
        etld_lib_credentials.main()
        credentials = etld_lib_credentials.get_cred()
        print(f"Qualys Login Test for {credentials.get('username')} "
              f"at api_fqdn_server: {credentials.get('api_fqdn_server')}\n")
        login_status = etld_lib_credentials.test_qualys_login_logout_basic_auth()
        if login_status['login_failed'] is not True:
            print(f"Testing Qualys Login/Logout for {credentials.get('username')} "
                  f"Succeeded at {credentials.get('api_fqdn_server')}\n"
                  f"    with HTTPS Return Code: {login_status['response_status_code']}.")
        etld_lib_functions.main(log_level=logging.WARN, my_logger_prog_name='qetl_manage_user')
        return True
    else:
        return False


def if_user_selected_test_qualys_basic_authentication_login_from_etld_lib_authentication_objects(first_time=False):
    if command_line_arguments.test is not None or first_time is not False:
        etld_lib_functions.main(log_level=logging.info, my_logger_prog_name='qetl_manage_user')
        etld_lib_config.main()
        #etld_lib_credentials.main()
        #credentials = etld_lib_credentials.get_cred()
        etld_lib_authentication_objects.main(test_about_qualys_flag=False)
        credentials = etld_lib_authentication_objects.qualys_authentication_obj.get_credentials()

        print(f"Qualys Login Test for {credentials.get('username')} "
              f"at api_fqdn_server: {credentials.get('api_fqdn_server')}\n")
        # login_status = etld_lib_credentials.test_qualys_login_logout_basic_auth()
        login_status_flag = etld_lib_authentication_objects.qualys_authentication_obj.test_qualys_login_logout_basic_auth()
        login_status = etld_lib_authentication_objects.qualys_authentication_obj.test_qualys_login_logout_basic_auth_result_dict
        credentials_source = etld_lib_authentication_objects.qualys_authentication_obj.credentials_source
        if login_status_flag is True:
            print(f"Testing Qualys Login/Logout for {credentials.get('username')} "
                  f"Succeeded at {credentials.get('api_fqdn_server')} with credentials_source: {credentials_source}\n"
                  f"    with HTTPS Return Code: {login_status['response_status_code']}.")
        else:
            print(f"Testing Qualys Login/Logout for {credentials.get('username')} "
                  f"Failed at {credentials.get('api_fqdn_server')} with credentials_source: {credentials_source}\n "
                  f"    with HTTPS Return Code: {login_status['response_status_code']} \n"
                  f"    Exception: {login_status['Exception']} \n"
                  f"    Message:   {login_status['message']}")

        etld_lib_functions.main(log_level=logging.WARN, my_logger_prog_name='qetl_manage_user')
        return True
    else:
        return False


# def if_user_selected_test_qualys_gateway_token_login(first_time=False):
#     if command_line_arguments.test is not None or first_time is not False:
#         etld_lib_functions.main(log_level=logging.INFO, my_logger_prog_name='qetl_manage_user')
#         etld_lib_config.main()
#         etld_lib_credentials.main()
#         credentials = etld_lib_credentials.get_cred()
#         print(f"Qualys Login Test for {credentials.get('username')} "
#               f"at gateway_fqdn_server: {credentials.get('gateway_fqdn_server')}\n")
#         etld_lib_credentials.test_gateway_auth()
#         if etld_lib_credentials.login_failed is not True:
#             print(f"Testing Qualys Login for {credentials.get('username')} "
#                   f"Succeeded at {credentials.get('api_fqdn_server')}\n"
#                   f"    with HTTPS Return Code: {etld_lib_credentials.http_return_code}.")
#         etld_lib_functions.main(log_level=logging.WARN, my_logger_prog_name='qetl_manage_user')


def if_user_selected_print_directory_listing_report():
    if command_line_arguments.report is not None:
        print(f"Report on user: {etld_lib_config.qetl_user_home_dir}")
        for path in sorted(etld_lib_config.qetl_user_home_dir.rglob('*')):
            depth = len(path.relative_to(etld_lib_config.qetl_user_home_dir).parts)
            spacer = '    ' * depth
            print(f'{spacer}+ {path.name}')
        print("\n")
        return True
    else:
        return False


def help_message(notes):
    help_mess = f'''
    {notes}
    usage: qetl_manage_user [-h] [-u qetl_USER_HOME_DIR] [-e etl_[module] ] [-e validate_etl_[module] ] [-c] [-t] [-i] [-d] [-r] [-l]
    
    Command to Extract, Transform and Load Qualys Data into various forms ( CSV, JSON, SQLITE3 DATABASE )
    
    optional arguments:
      -h, --help                show this help message and exit
      -u Home Directory Path,   --qetl_user_home_dir Home directory Path
                                   Example:
                                   - /opt/qetl/users/q_username
      -e etl_[module],          --execute_etl_[module] execute etl of module name. valid options are:
                                       -e etl_knowledgebase 
                                       -e etl_host_list 
                                       -e etl_host_list_detection
                                       -e etl_asset_inventory
                                       -e etl_was
                                       -e etl_pcrs
                                       -e etl_test_system ( for a small system test of all ETL Jobs )
      -e validate_etl_[module], --validate_etl_[module] [test last run of etl_[module]].  valid options are:
                                       -e validate_etl_knowledgebase
                                       -e validate_etl_host_list 
                                       -e validate_etl_host_list_detection
                                       -e validate_etl_asset_inventory
                                       -e validate_etl_was
                                       -e validate_etl_pcrs
                                       -e validate_etl_test_system 
      -d YYMMDDThh:mm:ssZ,      --datetime      YYYY-MM-DDThh:mm:ssZ UTC. Get All Data On or After Date. 
                                                Ex. 1970-01-01T00:00:00Z acts as flag to obtain all data.
      -c, --credentials        update qualys api user credentials: qualys username, password or api_fqdn_server
      -t, --test               test qualys credentials
      -i, --initialize_user    For automation, create a /opt/qetl/users/[userhome] directory 
                               without being prompted.
      -l, --logs               detailed logs sent to stdout for testing qualys credentials
      -v, --version            Help and QualysETL version information.
      -r, --report             brief report of the users directory structure.
      -p, --prompt-credentials prompt user for credentials, also accepts stdin with credentials piped to program.
      -m, --memory-credentials get credentials from environment: 
                               Example: q_username="your userid", q_password=your password, q_api_fqdn_server=api fqdn, q_gateway_fqdn_server=gateway api fqdn
      -s, --stdin-credentials  send credentials in json to stdin. 
                               Example:
                               {{"q_username": "your userid", "q_password": "your password", "q_api_fqdn_server": "api fqdn", "q_gateway_fqdn_server": "gateway api fqdn"}}
      
    Example: ETL Host List Detection
    
    qetl_manage_user -u [path] -e etl_host_list_detection -d 1970-01-01T00:00:00Z
    
     - qetl_manage_user will download all knowledgebase, host list and host list detection vulnerability data,
       transforming/loading it into sqlite and optionally the corresponding distribution directory.
     
     Inputs: 
       - KnowledgeBase API, Host List API, Host List Detection API.
       - ETL KnowledgeBase
         - /api/2.0/fo/knowledge_base/vuln/?action=list
       - ETL Host List
         - /api/2.0/fo/asset/host/?action=list
       - ETL Host List Detection - Stream of batches immediately ready for downstream database ingestion.
         - /api/2.0/fo/asset/host/vm/detection/?action=list
     Outputs:
       - XML, JSON, SQLITE, AND Distribution_Directory of CSV BATCH FILES PREPARED FOR DATABASE INGESTION.
         - host_list_detection_extract_dir - contains native xml and json transform of data from qualys, compressed in uniquely named batches.
         - host_list_detection_distribution_dir - contains transformed/prepared data ready for use in database loaders such as mysql.
         - host_list_detection_sqlite.db - sqlite database will contain multiple tables:
           - Q_Host_List                            - Host List Asset Data from Host List API.
           - Q_Host_List_Detection_Hosts            - Host List Asset Data from Host List Detection API. 
           - Q_Host_List_Detection_QIDS             - Host List Vulnerability Data from Host List Detection API. 
           - Q_KnowledgeBase_In_Host_List_Detection - KnowledgeBase QIDs found in Q_Host_List_Detection_QIDS. 
         
   etld_config_settings.yaml notes:
       - To Enable CSV Distribution, add the following keys to etld_config_settings.yaml and toggle on/off them via True or False
            kb_distribution_csv_flag: True                    # populates qetl_home/data/knowledgebase_distribution_dir
            host_list_distribution_csv_flag: True             # populates qetl_home/data/host_list_distribution_dir
            host_list_detection_distribution_csv_flag: True   # populates qetl_home/data/host_list_detection_distribution_dir
            asset_inventory_distribution_csv_flag: True       # populates qetl_home/data/asset_inventory_distribution_dir
            was_distribution_csv_flag: True                   # populates qetl_home/data/was_distribution_dir
              
            These files are prepared for database load, tested with mysql.  No headers are present.  
            Contact your Qualys TAM and schedule a call with David Gregory if you need assistance with this option.
            
      QualysETL Version: {qualys_etl.__version__}
    '''
    print(f"{help_mess}")


def test_command_line_arguments():
    global command_line_arguments
    # If no options check for -u.  If -u exists, help_message.  If -u does not exist, continue to new user prompts.
    # If -u is not set, help_message

    if command_line_arguments.version:
        help_message(notes="")
        exit(0)

    if command_line_arguments.qetl_user_home_dir is None:
        help_message(f"Please enter -u [ your /opt/qetl/users/ user home directory path ]\n\n"
                     f"    Note: /opt/qetl/users/newuser is the root directory for your qetl userhome directory,\n" 
                     f"    Example:\n"
                     f"             qetl_manage_user -u /opt/qetl/users/[your_user_name]\n"
                     f" ")
        exit(6)  # 6 is for testing the command works. Ex. qetl_manage_user; if [[ "$?" == "6"]]; then : ...

    test_qetl_user_home_dir = Path(command_line_arguments.qetl_user_home_dir).absolute()
    if test_qetl_user_home_dir.parent.parent.is_dir() and \
            os.access(str(test_qetl_user_home_dir.parent.parent), os.W_OK) and \
            test_qetl_user_home_dir.parent.name == 'users' and \
            test_qetl_user_home_dir.parent.parent.name == 'qetl' and \
            test_qetl_user_home_dir.parent.parent.parent.name == 'opt':
        pass
    else:
        help_message(f"Please check path and permissions on {test_qetl_user_home_dir.parent.parent},\n "
                     f"   You don't appear to have authorization to write to that directory.\n")
        exit(1)

    if test_qetl_user_home_dir.is_dir():
        if command_line_arguments.execute_etl_module is None and \
                command_line_arguments.credentials is None and \
                command_line_arguments.test is None and \
                command_line_arguments.logs is None and \
                command_line_arguments.initialize_user is None and \
                command_line_arguments.report is None:
            help_message(f"Please select an option for qetl_user: {command_line_arguments.qetl_user_home_dir}")
            exit(1)
    else:
        pass


def get_command_line_arguments(args=None):
    def dash_alpha_arg_greater_than_two_characters(arg):
        return len(arg) > 2 and arg[0] == '-' and arg[1].isalpha()
    global command_line_arguments
    parser = argparse.ArgumentParser(description='Command to Extract, Transform and Load Qualys Data into various forms ( CSV, JSON, SQLITE3 DATABASE )')
    parser.add_argument('-u', '--qetl_user_home_dir', default=None, help="Please enter -u option with path to your "
                                                                         "user directory.  Ex. /opt/qetl/users/[youruserdir]")
    parser.add_argument('-e', '--execute_etl_module', default=None,
                        help='etl_knowledgebase, etl_host_list, etl_host_list_detection, etl_asset_inventory, etl_was, etl_pcrs, '
                             'validate_etl_knowledgebase, validate_etl_host_list, validate_etl_host_list_detection, '
                             'validate_etl_was, validate_etl_pcrs, validate_etl_all, validate_etl_test_system')
    parser.add_argument('-d', '--datetime', default=None,
                        help='YYYY-MM-DDThh:mm:ssZ UTC. Get All Data On or After Date.  Ex. 1970-01-01T00:00:00Z acts as flag to obtain all data.')
    parser.add_argument('-c', '--credentials', default=None, action="store_true",
                        help='update qualys api user credentials stored on disk: qualys username, password or api_fqdn_server')
    parser.add_argument('-t', '--test', default=None, action="store_true", help='test qualys credentials')
    parser.add_argument('-i', '--initialize_user', default=None, action="store_true", help='create qualys user directory')
    parser.add_argument('-l', '--logs', default=None, action="store_true", help='detailed logs sent to stdout for test qualys credentials')
    parser.add_argument('-p', '--prompt_credentials', default=None, action="store_true", help='prompt user for credentials')
    parser.add_argument('-s', '--stdin_credentials', default=None, action="store_true",
                        help='read stdin credentials json {"q_username":"your userid", "q_password":"your password", "q_api_fqdn_server":"api fqdn", "q_gateway_fqdn_server":"gateway api fqdn"}')
    parser.add_argument('-m', '--memory_credentials', default=None, action="store_true",
                        help='Get credentials from environment variables in memory: q_username, q_password, q_api_fqdn_server, and optionally add q_gateway_fqdn_server. Ex. export q_username=myuser')
    parser.add_argument('-v', '--version', default=None, action="store_true", help='Help and QualysETL Version')
    parser.add_argument('-r', '--report', default=None, action="store_true", help='Brief report of the users directory structure.')
    # Print help menu from -v option.
    for idx, arg in enumerate(args):
        if '-h' in arg:
            args[idx] = '-v'
        if '--help' in arg:
            args[idx] = '-v'
        if dash_alpha_arg_greater_than_two_characters(arg):
            args[idx] = '-v'
    command_line_arguments = parser.parse_args(args)


def setup_qualys_etl_user_home_environment_in_memory_and_in_etld_lib_config():
    global qetl_user_home
    global command_line_arguments
    # Reset Logging.
    if command_line_arguments.execute_etl_module is None:
        if command_line_arguments.logs is None:
            etld_lib_functions.main(log_level=logging.WARNING, my_logger_prog_name='qetl_manage_user')
        else:
            etld_lib_functions.main(log_level=logging.INFO, my_logger_prog_name='qetl_manage_user')
    else:
        pass

    # qetl_user_home_dir
    os.environ['qualys_etl_user_home'] = command_line_arguments.qetl_user_home_dir
    etld_lib_config.set_path_qetl_user_home_dir()  # If qetl_user_home_dir is malformed, we abort here.
    if etld_lib_config.qetl_user_home_dir.is_dir():
        # Directory Exists.  Options are test, report, execute, etc...
        pass
    else:
        # Potential New User, Query for confirmation
        time.sleep(1)
        qetl_manage_user_init=""
        if command_line_arguments.initialize_user is not None:
            print(f"\nqetl_user_home_dir does not exist: {etld_lib_config.qetl_user_home_dir}")
            response = 'qetl_manage_user_init'
        else:
            print(f"\nqetl_user_home_dir does not exist: {etld_lib_config.qetl_user_home_dir}")
            response = input(f"Create new qetl_user_home_dir? {etld_lib_config.qetl_user_home_dir} ( yes or no ): ")

        if response == 'yes':
            if command_line_arguments.execute_etl_module is not None:
                if command_line_arguments.logs is None:
                    etld_lib_functions.main(log_level=logging.WARNING, my_logger_prog_name='qetl_manage_user')
                else:
                    etld_lib_functions.main(log_level=logging.INFO, my_logger_prog_name='qetl_manage_user')
            etld_lib_config.qetl_create_user_dirs_ok_flag = True
            etld_lib_config.main()
            time.sleep(1)
            print(f"\nqetl_user_home_dir created: {etld_lib_config.qetl_user_home_dir}")
            #if_user_selected_update_credentials_stored_on_disk(first_time=True)
            if_user_selected_update_credentials_stored_on_disk_from_etld_lib_authentication_objects(first_time=True)
            response = input(f"\nWould you like to test login/logout of Qualys API? ( yes or no ): ")
            if response == 'yes':
                print("")
                #if_user_selected_test_qualys_basic_authentication_login(first_time=True)
                if_user_selected_test_qualys_basic_authentication_login_from_etld_lib_authentication_objects(first_time=True)
            print(f"\nThank you, exiting.\n")
            exit(0)
        elif response == 'qetl_manage_user_init':
            etld_lib_config.qetl_create_user_dirs_ok_flag = True
            etld_lib_functions.main(log_level=logging.WARNING, my_logger_prog_name='qetl_manage_user_init')
            etld_lib_config.main()
            time.sleep(1)
            print(f"\nqetl_user_home_dir created: {etld_lib_config.qetl_user_home_dir}")
            print(f"Thank you, exiting.\n")
            exit(0)
        else:
            print(f"\nThank you, exiting.\n")
            exit(1)


def if_root_user_abort_program(logger_method=print):
    userid = int(os.getuid())
    if userid == 0:
        logger_method(f"Sorry, please run as non-root user.  Current userid is: {userid}")
        logger_method(f"Exiting program - Cannot run qetl_manage_user as root user.")
        exit(1)


def main():
    if_root_user_abort_program()
    get_command_line_arguments(sys.argv[1:])
    test_command_line_arguments()
    setup_qualys_etl_user_home_environment_in_memory_and_in_etld_lib_config()

    if if_user_selected_sending_credentials_via_stdin():
        pass
    elif if_prompt_credentials():
        pass
    elif if_user_selected_setting_credentials_in_memory():
        pass
    elif if_user_selected_update_credentials_stored_on_disk_from_etld_lib_authentication_objects():
        pass
    #elif if_user_selected_update_credentials_stored_on_disk():
    #    pass

    #if_user_selected_test_qualys_basic_authentication_login()
    if_user_selected_test_qualys_basic_authentication_login_from_etld_lib_authentication_objects()

    if if_user_selected_print_directory_listing_report():
        pass
    else:
        if_user_selected_execute_an_etl_module()


if __name__ == '__main__':
    main()






import base64
import time
import requests
from pathlib import Path
import shutil
import yaml
# 2023-05-13 import oschmod
import re
import os
import stat
from urllib.parse import urlencode, quote_plus
from qualys_etl.etld_lib import etld_lib_functions
from qualys_etl.etld_lib import etld_lib_config
from qualys_etl.etld_lib import etld_lib_datetime
from qualys_etl.etld_lib import etld_lib_timer_obj
from qualys_etl.etld_lib import etld_lib_oschmod as oschmod
from dataclasses import dataclass
import qualys_etl

global qualys_authentication_obj
#global qualys_basic_authorization_obj
#global qualys_bearer_token_obj

@dataclass
class QualysAuthenticationObj:

    def __init__(self, username="", password="", api_fqdn_server="", gateway_fqdn_server="",
                 acquire_bearer_token=False, max_bearer_token_age=900, test_basic_authorization=False,
                 initialize_etld_cred_yaml_flag=True,
                 logger_info=print, logger_error=print, logger_warning=print,
                 http_user_agent_etl_workflow_message="auth_obj", test_about_qualys_flag=True):
        self.username = username
        self.password = password
        self.api_fqdn_server = api_fqdn_server
        self.gateway_fqdn_server = gateway_fqdn_server
        self.authorization = ""
        self.bearer = ""
        self.bearer_timer_obj : etld_lib_timer_obj.TimerObj = ""
        self.bearer_token_required = acquire_bearer_token
        self.max_bearer_token_age = max_bearer_token_age
        self.get_bearer_token_from_qualys_succeeded = False
        self.test_authorization_result = False
        self.test_about_qualys_flag = test_about_qualys_flag
        self.test_basic_authorization = test_basic_authorization
        self.initialize_etld_cred_yaml_flag = initialize_etld_cred_yaml_flag
        self.logger_info = logger_info
        self.logger_error = logger_error
        self.logger_warning = logger_warning
        self.http_user_agent_etl_workflow_message = http_user_agent_etl_workflow_message
        self.credentials_source = ""
        self.set_credentials()
        self._set_authorization()
        self.test_qualys_login_logout_basic_auth_result_dict = ""
        if self.test_about_qualys_flag:
            self.test_about_qualys()
        if self.bearer_token_required:
            self.bearer = self.get_current_bearer_token()
        if self.test_basic_authorization:
            self.test_authorization_result = \
                self.test_qualys_login_logout_basic_auth(
                    username=self.username, password=self.password, api_fqdn_server=self.api_fqdn_server)

    def get_authorization(self) -> str:
        if self.authorization == "":
            self._set_authorization()
        return self.authorization

    def _set_authorization(self, username="", password=""):

        def get_base64_encoding(username="", password=""):
            auth_base64 = ""
            if username != "" and password != "":
                auth_base64 = 'Basic ' + base64.b64encode(f"{username}:{password}".encode('utf-8')).decode('utf-8')
            return auth_base64

        if username != "" and password != "":
            self.authorization = get_base64_encoding(username, password)
        elif self.username != "" and self.password != "":
            self.authorization = get_base64_encoding(self.username, self.password)
        else:
            self.authorization = ""

    def _get_gateway_fqdn_server(self):
        if self.gateway_fqdn_server == "":
            self._set_gateway_fqdn_server()
        return self.gateway_fqdn_server

    def _set_gateway_fqdn_server(self, api_fqdn_server="", gateway_fqdn_server=""):
        def get_gateway_platform_fqdn(api_fqdn_server) -> str:
            platform_url = self.get_platform_url_dict()
            gateway_fqdn_server = ""
            if api_fqdn_server in platform_url.keys():
                gateway_fqdn_server = platform_url[api_fqdn_server]
            return gateway_fqdn_server

        if gateway_fqdn_server != "":
            self.gateway_fqdn_server = gateway_fqdn_server
        elif api_fqdn_server != "":
            self.gateway_fqdn_server = get_gateway_platform_fqdn(api_fqdn_server)
        else:
            self.gateway_fqdn_server = get_gateway_platform_fqdn(self.api_fqdn_server)

    def get_credentials(self) -> dict:
        cred = {'username': self.username, 'password': self.password,
                'api_fqdn_server': self.api_fqdn_server, 'gateway_fqdn_server': self.gateway_fqdn_server}
        return cred

    def set_credentials(self, force_from_file=False):

        def initialize_eltd_cred_yaml_file(cred_file: Path = None):
            if not Path.is_file(cred_file):
                etld_lib_config.set_qetl_code_dir_etld_cred_yaml_template_path()
                cred_example_file_path = Path(etld_lib_config.qetl_code_dir_etld_cred_yaml_template_path)
                #cred_example_file_path = Path(etld_lib_config.set_qetl_code_dir_etld_cred_yaml_template_path())
                destination_file_path = Path(etld_lib_config.qetl_user_cred_file)
                shutil.copy(str(cred_example_file_path), str(destination_file_path), follow_symlinks=True)
                cred_example_file = open(str(cred_example_file_path), "r", encoding='utf-8')
                cred_example = cred_example_file.read()
                cred_example_file.close()
                local_date = etld_lib_datetime.get_local_date()  # Add date updated to file
                cred_example = re.sub('\$DATE', local_date, cred_example)
                cred_file_example = open(str(cred_file), 'w', encoding='utf-8')
                cred_file_example.write(cred_example)
                cred_file_example.close()
            oschmod.set_mode(str(cred_file), "u+rw,u-x,go-rwx")

        def set_credentials_from_environment(username="", password="", api_fqdn_server="", gateway_fqdn_server=""):
            k = os.environ.keys()
            if 'q_username' in k:
                username = os.environ.get('q_username')
            if 'q_password' in k:
                password = os.environ.get('q_password')
            if 'q_api_fqdn_server' in k:
                api_fqdn_server = os.environ.get('q_api_fqdn_server')
            if 'q_gateway_fqdn_server' in k:
                gateway_fqdn_server = os.environ.get('q_gateway_fqdn_server')

            if username != "" and password != "" and api_fqdn_server != "":
                self.username = username
                self.password = password
                self.api_fqdn_server = api_fqdn_server
                self.gateway_fqdn_server = gateway_fqdn_server
                self.credentials_source = 'set_credentials_from_environment'
                self.logger_info("set_credentials_from_environment: "
                                               f"username: {username}, "
                                               f"api_fqdn_server:  {api_fqdn_server}, "
                                               f"gateway_fqdn_server={gateway_fqdn_server}")

        def set_credentials_from_yaml_file(cred_file=""):
            try:
                with open(cred_file, 'r', encoding='utf-8') as cred_yaml_file:
                    cred = yaml.safe_load(cred_yaml_file)
                    if 'username' in cred.keys() and 'password' in cred.keys() and 'api_fqdn_server' in cred.keys():
                        username = cred['username']
                        password = cred['password']
                        api_fqdn_server = cred['api_fqdn_server']
                    else:
                        self.logger_error(f"cred yaml contents: {cred}")
                        raise Exception(f"credentials not in file: {cred_file}")

                    if 'gateway_fqdn_server' in cred.keys():
                        gateway_fqdn_server = cred['gateway_fqdn_server']
                    else:
                        self._set_gateway_fqdn_server(api_fqdn_server=api_fqdn_server)
                        gateway_fqdn_server = self._get_gateway_fqdn_server()

                    self.username = username
                    self.password = password
                    self.api_fqdn_server = api_fqdn_server
                    self.gateway_fqdn_server = gateway_fqdn_server
                    self.credentials_source = 'set_credentials_from_yaml_file'
                    self.logger_info("set_credentials_from_yaml_file: "
                                                       f"username: {username}, "
                                                       f"api_fqdn_server:  {api_fqdn_server}, "
                                                       f"gateway_fqdn_server={gateway_fqdn_server}")

            except Exception as e:
                self.logger_error(f"Please run qetl_manage_user -c -u [yourdir] to initialize your credentials file")
                self.logger_error(f"Exception: {e}")
                raise Exception(f"Exception {e}")

        def set_credentials_from_vault(vault_type='hashicorp'):
            # Holder for vault
            pass

        #
        # MAIN ROUTINE
        #

        if self.initialize_etld_cred_yaml_flag:
            initialize_eltd_cred_yaml_file(cred_file=etld_lib_config.qetl_user_cred_file)

        obtain_credentials_from_vault = False
        if force_from_file:
            set_credentials_from_yaml_file(cred_file=etld_lib_config.qetl_user_cred_file)
        elif self.username != "" or self.api_fqdn_server != "" or self.gateway_fqdn_server != "" or self.password != "":
            self.credentials_source = 'Testing - QualysAuthenticationObject __init__ args'
            self.logger_info("set_credentials_at_init_options: "
                             f"username: {self.username}, "
                             f"api_fqdn_server:  {self.api_fqdn_server}, "
                             f"gateway_fqdn_server={self.gateway_fqdn_server}")
        elif 'q_username' in os.environ.keys() \
                and 'q_password' in os.environ.keys() \
                and 'q_api_fqdn_server' in os.environ.keys():
            set_credentials_from_environment()
        elif obtain_credentials_from_vault:
            # Add method to obtain credentials from vault
            pass
        else:
            set_credentials_from_yaml_file(cred_file=etld_lib_config.qetl_user_cred_file)

        if self.username == "" or self.password == "" or self.api_fqdn_server == "" or self.credentials_source == "":
            self.logger_error("Could not set credentials. Please ensure credentials are set via stdin, environment or in .etld_cred.yaml.")
            exit(1)

    def update_yaml_file_credentials(self, new_cred):
        # Get Current .etld_cred.yaml file
        with open(etld_lib_config.qetl_user_cred_file, 'r', encoding='utf-8') as cred_yaml_file:
            current_cred = yaml.safe_load(cred_yaml_file)
        # Get Template
        cred_example_file_path = Path(etld_lib_config.qetl_code_dir_etld_cred_yaml_template_path)
        with open(str(cred_example_file_path), "r", encoding='utf-8') as cred_template_file:
            cred_template_string = cred_template_file.read()
        # Update Template # username: initialuser  password: initialpassword  api_fqdn_server: qualysapi.qualys.com
        if current_cred == new_cred:
            pass
        else:
            if new_cred['username'] == current_cred['username']:
                new_username = f"username: '{current_cred.get('username')}'"
            else:
                new_username = f"username: '{new_cred.get('username')}'"

            if new_cred['password'] == current_cred['password']:
                new_password = f"password: '{current_cred.get('password')}'"
            else:
                new_password = f"password: '{new_cred.get('password')}'"

            if new_cred['api_fqdn_server'] == current_cred['api_fqdn_server']:
                new_api_fqdn_server = f"api_fqdn_server: '{current_cred.get('api_fqdn_server')}'"
            else:
                new_api_fqdn_server = f"api_fqdn_server: '{new_cred.get('api_fqdn_server')}'"

            if new_cred['gateway_fqdn_server'] == current_cred['gateway_fqdn_server']:
                new_gateway_fqdn_server = f"gateway_fqdn_server: '{current_cred.get('gateway_fqdn_server')}'"
            else:
                new_gateway_fqdn_server = f"gateway_fqdn_server: '{new_cred.get('gateway_fqdn_server')}'"

            local_date = etld_lib_datetime.get_local_date()
            cred_template_string = re.sub('\$DATE', local_date, cred_template_string)
            cred_template_string = re.sub('username: initialuser', new_username, cred_template_string)
            # cred_template_string = re.sub('password: initialpassword', new_password, cred_template_string)
            cred_template_string = cred_template_string.replace('password: initialpassword', new_password)
            cred_template_string = re.sub('api_fqdn_server: qualysapi.qualys.com', new_api_fqdn_server,
                                          cred_template_string)
            cred_template_string = re.sub('gateway_fqdn_server: gateway.qg1.apps.qualys.com', new_gateway_fqdn_server,
                                          cred_template_string)
            with open(etld_lib_config.qetl_user_cred_file, 'w', encoding='utf-8') as cred_file_to_update:
                cred_file_to_update.write(cred_template_string)
            oschmod.set_mode(etld_lib_config.qetl_user_cred_file, "u+rw,u-x,go-rwx")

    def get_current_bearer_token(self, force_update_to_bearer_token=False) -> str:
        if force_update_to_bearer_token:
            self.bearer = ""
            self.bearer_timer_obj = ""

        if self.bearer == "":
            self.bearer = self._get_bearer_token_from_qualys()
            if self.bearer == "":
                self.get_bearer_token_from_qualys_succeeded = False
            else:
                self.get_bearer_token_from_qualys_succeeded = True

        if self.bearer_timer_obj == "" and self.get_bearer_token_from_qualys_succeeded:
            self.bearer_timer_obj = etld_lib_timer_obj.TimerObj(max_time=self.max_bearer_token_age)

        if self.get_bearer_token_from_qualys_succeeded:
            max_time_exceeded_flag = \
                self.bearer_timer_obj.if_time_elapsed_past_max_time(max_time=self.max_bearer_token_age)
            if max_time_exceeded_flag:
                self.bearer = self._get_bearer_token_from_qualys(
                    gateway_fqdn_server=self.gateway_fqdn_server,
                    username=self.username,
                    password=self.password
                )
                if self.bearer == "":
                    self.get_bearer_token_from_qualys_succeeded = False
                else:
                    self.get_bearer_token_from_qualys_succeeded = True
                    self.bearer_timer_obj.reset_timer()

        return self.bearer

    def _get_bearer_token_from_qualys(self, gateway_fqdn_server="", username="", password="") -> str:

        bearer_token = ""
        if gateway_fqdn_server == "":
            gateway_fqdn_server = self.gateway_fqdn_server
            if gateway_fqdn_server == "":
                self._set_gateway_fqdn_server(api_fqdn_server=self.api_fqdn_server)
        if gateway_fqdn_server == "" or username == "" or password == "":
            gateway_fqdn_server = self.gateway_fqdn_server
            username = self.username
            password = self.password

        if gateway_fqdn_server == "" or username == "" or password == "":
            self.logger_error(f"Please add gateway_fqdn_server credentials file or environment.  "
                                            f"gateway: {gateway_fqdn_server}, username: {username}")
            return bearer_token

        url = f"https://{gateway_fqdn_server}/auth"  # Qualys Endpoint
        payload = {'token': 'true', 'password': password, 'username': username, 'permissions': 'true'}
        payload = urlencode(payload, quote_via=quote_plus)

        headers = {'X-Requested-With': f'qualysetl_v{qualys_etl.__version__} {self.http_user_agent_etl_workflow_message}',
                   'Content-Type': 'application/x-www-form-urlencoded',
                   'User-Agent': f"qualysetl_v{qualys_etl.__version__} {self.http_user_agent_etl_workflow_message}"}

        max_retries = 5
        sleep_time_requests_module_failed = 300
        sleep_time_503 = 30
        retry_number = 0
        for retry_number in range(max_retries):
            try:
                response = requests.request("POST", url, headers=headers, data=payload,
                                            verify=etld_lib_config.requests_module_tls_verify_status)
                if response.status_code == 503 or response.status_code == 500:  # 503/500:  # Service Temporarily Unavailable
                    self.gateway_log_message(
                        response=response,
                        message=f"Gateway Login Failed - Retry number {retry_number} for HTTP Response Code: {response.status_code} ",
                        log_level=self.logger_warning)
                    time.sleep(sleep_time_503)
                elif response.status_code == 201: # Success
                    bearer_token = f"Bearer {response.text}"
                    if self.bearer_timer_obj == "":
                       bearer_timer_obj_message = "bearer_timer_obj= NOT SET"
                    else:
                        bearer_timer_obj_message = f"bearer_timer_obj={self.bearer_timer_obj.__dict__}"
                    self.gateway_log_message(
                        response=response,
                        message=f"Gateway Login Success - Bearer Token Reset - HTTP Response Code: {response.status_code}, max_bearer_token_age={self.max_bearer_token_age}, {bearer_timer_obj_message} ",
                        log_level=self.logger_info)
                    break
                else:
                    self.gateway_log_message(
                        response=response,
                        message=f"Gateway Login Failed - HTTP Response Code: {response.status_code} ",
                        log_level=self.logger_error)
                    break
            except requests.exceptions.RequestException as e:
                # Timeout Example
                self.gateway_log_message(
                    response="",
                    message=f"Gateway Login Failed - Exception {e}",
                    log_level=self.logger_warning)
                time.sleep(sleep_time_requests_module_failed)
        else:
            self.gateway_log_message(
                response="ALL RETRIES FAILED",
                message=f"Gateway Login Failed - Retry number {retry_number} of {max_retries}",
                log_level=self.logger_error)

        return bearer_token

    def test_about_qualys(self):
        api_fqdn_server = self.api_fqdn_server
        authorization = self.authorization

        url = f"https://{api_fqdn_server}/msp/about.php"  # Qualys Endpoint
        headers = {
            'X-Requested-With': f'qualysetl_v{qualys_etl.__version__} {self.http_user_agent_etl_workflow_message}',
            'User-Agent': f"qualysetl_v{qualys_etl.__version__} {self.http_user_agent_etl_workflow_message}",
            'Authorization': authorization}
        payload = {}

        try:
            response = requests.request("POST", url, headers=headers, data=payload,
                                        verify=etld_lib_config.requests_module_tls_verify_status)

            if response.status_code == 200:
                response_text = str(response.text).replace('\n', ' ')
                self.logger_info(f"msp/about http_response_status_code={response.status_code}, "
                                 f"response_text={response_text}")
            else:
                response_text = str(response.text).replace('\n', ' ')
                self.logger_info(f"msp/about http_response_status_code={response.status_code}, "
                                 f"response_text={response_text}")

        except Exception as e:
            self.logger_warning(f"msp/about failed with exception: {e}")

    def test_qualys_login_logout_basic_auth(self, api_fqdn_server="", username="", password="") -> bool:

        def test_qualys_logout(login_status_dict=None) -> dict:
            logout_status_dict = {'test_response': False, 'response_status_code': "", 'Exception': "", 'message': ""}
            if login_status_dict is None:
                logout_status_dict['test_response'] = True
                logout_status_dict['message'] = "No login_status_dict, continue without logout"
            elif login_status_dict['cookie'] == "":
                logout_status_dict['test_response'] = True
                logout_status_dict['message'] = "No login_status_dict['cookie'], continue without logout"
            else:
                url = f"https://{api_fqdn_server}/api/2.0/fo/session/"  # Qualys Endpoint
                payload = {'action': 'logout'}
                cookie = login_status_dict['cookie']
                headers = {'X-Requested-With': f'qualysetl_v{qualys_etl.__version__} {self.http_user_agent_etl_workflow_message}',
                           'Content-Type': 'application/x-www-form-urlencoded', 'Cookie': cookie,
                           'User-Agent': f"qualysetl_v{qualys_etl.__version__} {self.http_user_agent_etl_workflow_message}"}
                try:
                    response = requests.request("POST", url, headers=headers, data=payload,
                                                verify=etld_lib_config.requests_module_tls_verify_status)
                    logout_status_dict['response_status_code'] = str(response.status_code)
                    if response.status_code == 200:
                        logout_status_dict['test_response'] = True
                except Exception as e:
                    logout_status_dict['test_response'] = False
                    logout_status_dict['message'] = "Logout with Exception, continue without logout"
                    logout_status_dict['response_status_code'] = ""
                    logout_status_dict['Exception'] = e

            return logout_status_dict

        def get_cookie_from_response(response):
            cookie_dict = response.cookies.get_dict()
            cookie = f"DWRSESSIONID={cookie_dict['DWRSESSIONID']}; QualysSession={cookie_dict['QualysSession']}"
            cookie = cookie.replace('\n', '').replace('\r', '')
            return cookie

        def test_qualys_login(username="", password="", api_fqdn_server=""):
            login_status_dict = \
                {'test_response': False, 'response_status_code': "",
                 'Exception': "", 'message': "", 'cookie': ""}

            url = f"https://{api_fqdn_server}/api/2.0/fo/session/"
            payload = {'action': 'login', 'username': username, 'password': password}
            payload = urlencode(payload, quote_via=quote_plus)

            headers = {'X-Requested-With': f'qualysetl_v{qualys_etl.__version__} {self.http_user_agent_etl_workflow_message}',
                       'Content-Type': 'application/x-www-form-urlencoded',
                       'User-Agent': f"qualysetl_v{qualys_etl.__version__} {self.http_user_agent_etl_workflow_message}"}

            try:
                response = requests.request("POST", url, headers=headers, data=payload,
                                            verify=etld_lib_config.requests_module_tls_verify_status)
                login_status_dict['response_status_code'] = str(response.status_code)
                if response.status_code == 200:
                    cookie = get_cookie_from_response(response)
                    login_status_dict['cookie'] = cookie
                    login_status_dict['test_response'] = True
                elif response.status_code == 401:
                    login_status_dict['message'] = 'Invalid userid/password or api_fqdn_servers.  ' \
                                                   'Please check your credentials and rerun.'
                elif str(response.status_code).startswith('5'):
                    login_status_dict['message'] = 'Please check if Qualys is in a maintenance window, and rerun.'
            except requests.exceptions.RequestException as e:
                login_status_dict['message'] = 'API Exception, determine if you have connectivity issues and rerun.'
                login_status_dict['Exception'] = e

            return login_status_dict

        if username == "" or password == "" or api_fqdn_server == "":
            username = self.username
            password = self.password
            api_fqdn_server = self.api_fqdn_server

        self.test_qualys_login_logout_basic_auth_result_dict = {}
        login_status: dict = test_qualys_login(
            username=username, password=password, api_fqdn_server=api_fqdn_server)
        self.test_qualys_login_logout_basic_auth_result_dict = login_status
        login_test_flag = False

        if login_status['test_response']:
            self.logger_info(f"LOGIN  - Qualys Success with user: {username}")
            logout_status: dict = test_qualys_logout(login_status)
            login_test_flag = True
            if logout_status['test_response']:
                self.logger_info(f"LOGOUT - Qualys Success with user: {username}")
            else:
                self.logger_warning(f"LOGOUT - Qualys did not work for user: {username}")
                self.logger_warning(f"LOGOUT - Response:  {logout_status['response_status_code']}")
                self.logger_warning(f"LOGOUT - Message:   {logout_status['message']}")
                self.logger_warning(f"LOGOUT - Exception: {logout_status['Exception']}")
        else:
            self.logger_error(f"LOGIN - Qualys did not work for user: {username}")
            self.logger_error(f"LOGIN - Response:  {login_status['response_status_code']}")
            self.logger_error(f"LOGIN - Message:   {login_status['message']}")
            self.logger_error(f"LOGIN - Exception: {login_status['Exception']}")

        return login_test_flag

    def gateway_log_message(self, response, log_level, message=""):
        if response != "":
            response_message = f", HTTP RESPONSE CODE: {response.status_code}"
        else:
            response_message = ""

        log_level(f"{message} - username={self.username}, "
                  f"api_fqdn_server={self.api_fqdn_server}, "
                  f"gateway_fqdn_server={self.gateway_fqdn_server}")

    def get_platform_url_dict(self) -> dict:
        # https://www.qualys.com/platform-identification/
        platform_url = {
            'qualysapi.qualys.com': 'gateway.qg1.apps.qualys.com',
            'qualysapi.qg2.apps.qualys.com': 'gateway.qg2.apps.qualys.com',
            'qualysapi.qg3.apps.qualys.com': 'gateway.qg3.apps.qualys.com',
            'qualysapi.qg4.apps.qualys.com': 'gateway.qg4.apps.qualys.com',
            'qualysapi.qualys.eu': 'gateway.qg1.apps.qualys.eu',
            'qualysapi.qg2.apps.qualys.eu': 'gateway.qg2.apps.qualys.eu',
            'qualysapi.qg1.apps.qualys.in': 'gateway.qg1.apps.qualys.in',
            'qualysapi.qg1.apps.qualys.ca': 'gateway.qg1.apps.qualys.ca',
            'qualysapi.qg1.apps.qualys.ae': 'gateway.qg1.apps.qualys.ae',
            'qualysapi.qg1.apps.qualys.co.uk': 'gateway.qg1.apps.qualys.co.uk',
            'qualysapi.qg1.apps.qualys.com.au': 'gateway.qg1.apps.qualys.com.au',
            'qualysapi.qg1.apps.qualysksa.com': 'gateway.qg1.apps.qualysksa.com',
        }
        return platform_url

    def get_gateway_platform_fqdn(self, api_fqdn_server) -> str:
        platform_url = self.get_platform_url_dict()
        gateway_fqdn_server = ""
        if api_fqdn_server in platform_url.keys():
            gateway_fqdn_server = platform_url[api_fqdn_server]
        return gateway_fqdn_server

    def get_credentials_dict(self) -> dict:
        cred = {}
        cred['username'] = self.username
        cred['password'] = self.password
        cred['api_fqdn_server'] = self.api_fqdn_server
        cred['gateway_fqdn_server'] = self.gateway_fqdn_server
        cred['authorization'] = self.get_authorization()
        cred['bearer'] = self.bearer
        return cred


def main(**qualys_authentication_obj_arguments):
    global qualys_authentication_obj
    qualys_authentication_obj = QualysAuthenticationObj(
        logger_info=etld_lib_functions.logger.info,
        logger_error=etld_lib_functions.logger.error,
        logger_warning=etld_lib_functions.logger.warning,
        **qualys_authentication_obj_arguments
    )


def main_no_logger_for_qetl_manage_user(**qualys_authentication_obj_arguments):
    global qualys_authentication_obj
    qualys_authentication_obj = QualysAuthenticationObj(**qualys_authentication_obj_arguments)


def test_qualys_authentication_obj(**main_arguments):
    main(**main_arguments)
    qualys_authentication_obj.test_qualys_login_logout_basic_auth()
    bearer_token_test = []
    jwt_token_previous = qualys_authentication_obj.get_current_bearer_token()
    for time_in_seconds in [2, 3, 4, 6, 5, 7]:
        jwt_token = qualys_authentication_obj.get_current_bearer_token()
        if jwt_token == jwt_token_previous:
            pass
        else:
            bearer_token_test.append(jwt_token)
            jwt_token_previous = jwt_token
        time.sleep(time_in_seconds)

    count = 0
    for jwt_token in bearer_token_test:
        etld_lib_functions.logger.info(f"TOKEN {count} - JWT = {str(jwt_token)[-50:]}")
        count = count + 1

    if len(bearer_token_test) == 3:
        etld_lib_functions.logger.info(f"Pass JWT Token Test, token updated 3 times at 0, 6 and 7")
    else:
        etld_lib_functions.logger.error(f"Invalid JWT Token Updates.  Should be updated only 3 times at 0, 6 and 7")

if __name__ == '__main__':
    etld_lib_functions.main()
    etld_lib_config.main()
    test_qualys_authentication_obj(max_bearer_token_age=5)



import fcntl
import logging
import multiprocessing
import shutil
import sys
import time
import re
from pathlib import Path
from datetime import datetime
import timeit
from contextlib import redirect_stdout, redirect_stderr

from qualys_etl.etld_lib import etld_lib_config
from qualys_etl.etld_lib import etld_lib_credentials
from qualys_etl.etld_lib import etld_lib_authentication_objects
from qualys_etl.etld_lib import etld_lib_functions
from qualys_etl.etld_lib import etld_lib_log_validation

global target_module_to_run
global target_method_in_module_to_run
global spawned_process
global spawned_process_max_run_time
global spawned_process_start_time
global spawned_process_stop_time
global spawned_process_sleep_time
global spawned_process_count_to_status_update
global spawned_process_status_count
global log_file_max_size
global log_file_path
global log_file_rotate_path
global lock_file
global log_dir
global etl_dir_to_monitor
global etl_files_mtime_dict
global etl_files_max_time_between_changes
global etl_file_stalled


def spawn_etl_in_background_with_arg(args_list=[], max_proc=None):
    global spawned_process
    global spawned_process_start_time
    global spawned_process_status_count
    spawn_process_list = []

    # target=etl_host_list_detection
    # name=etl_hld_d_2021_06_03T00:00:00Z_b_000001
    # args=list_of_host_ids
    # Create jobs list  ready to spawn
    for arg in args_list:
        arg: dict
        arg.get('target_module_to_run')
        arg.get('target_method_in_module_to_run')
        arg.get('method_arguments')

        spawned_process = multiprocessing.Process(
            target=arg.get('target_module_to_run'),
            name=arg.get('target_method_in_module_to_run'),
            args=arg.get('method_arguments'))
        spawn_process_list.append(spawned_process)


def spawn_etl_in_background():
    global spawned_process
    global spawned_process_start_time
    global spawned_process_status_count
    global etl_dir_to_monitor
    global etl_files_mtime_dict
    etl_files_mtime_dict = {}

    spawned_process = multiprocessing.Process(target=target_module_to_run, name=target_method_in_module_to_run)
    spawned_process.start()
    etld_lib_functions.logger.info("Spawned ETL in Background")
    spawned_process_start_time = timeit.default_timer()
    spawned_process_status_count = 0
    etl_initial_file_for_monitoring = Path(etl_dir_to_monitor, str(target_method_in_module_to_run))
    Path.touch(etl_initial_file_for_monitoring, exist_ok=True)

    while spawned_process.is_alive():
        spawned_process_report_status()

    spawned_process_ended()
    time.sleep(1)


def etl_files_are_stalled():
    global etl_dir_to_monitor
    global etl_files_max_time_between_changes
    global etl_file_stalled
    etl_files_max_time_exceeded_flag = True
    etl_files = Path(etl_dir_to_monitor).glob('**/*')
    vm_workflow = \
        re.fullmatch(r"^.*(knowledgebase_etl_workflow|host_list_etl_workflow|host_list_detection_etl_workflow).*$",
                     str(target_method_in_module_to_run))
    asset_inventory_workflow = \
        re.fullmatch(r"^.*(asset_inventory_etl_workflow).*$",
                     str(target_method_in_module_to_run))

    was_workflow = \
        re.fullmatch(r"^.*(was_etl_workflow).*$",
                     str(target_method_in_module_to_run))

    pcrs_workflow = \
        re.fullmatch(r"^.*(pcrs_etl_workflow).*$",
                     str(target_method_in_module_to_run))

    test_system_workflow = \
        re.fullmatch(
            r"^.*(test_system_etl_workflow).*$", str(target_method_in_module_to_run))

    try:
        if vm_workflow:
            workflow_pattern = re.compile(r"^.*(/knowledgebase|/kb|/host_list).*$")
        elif asset_inventory_workflow:
            workflow_pattern = re.compile(r"^.*(/asset_inventory).*$")
        elif was_workflow:
            workflow_pattern = re.compile(r"^.*(/was).*$")
        elif pcrs_workflow:
            workflow_pattern = re.compile(r"^.*(/pcrs).*$")
        elif test_system_workflow:
            workflow_pattern = re.compile(r"^.*(/test_system|/knowledgebase|/kb|/host_list|/asset_inventory|/was|/pcrs).*$")
        else:
            etld_lib_functions.logger.error("Could not determine workflow etl_workflow_validation_type.")
            etld_lib_functions.logger.error("Please investigate.")
            raise Exception("Error determining workflow etl_workflow_validation_type")

        for etl_file_path in etl_files:
            if Path(etl_file_path).is_file():
                workflow_pattern_match = re.fullmatch(workflow_pattern, str(Path(etl_file_path)))
                if workflow_pattern_match:
                    mtime = int(Path(etl_file_path).stat().st_mtime)
                    datetime_now = datetime.now().timestamp()
                    datetime_diff = datetime_now - mtime
                    if datetime_diff < etl_files_max_time_between_changes:
                        etl_files_max_time_exceeded_flag = False
                        break  # Files have changed, break and continue.
                    else:
                        etl_file_stalled = Path(etl_file_path)
    except Exception as e:
        etl_files_max_time_exceeded_flag = True
        etld_lib_functions.logger.error(f"Could not determine file paths for monitoring: {etl_dir_to_monitor}")
        etld_lib_functions.logger.error(f"Please investigate Error: {e}.")

    if etl_files_max_time_exceeded_flag:
        etld_lib_functions.logger.info(f"{str(target_method_in_module_to_run)} is stalled.")

    return etl_files_max_time_exceeded_flag


def spawned_process_report_status():
    now = timeit.default_timer()
    run_time = (now - spawned_process_start_time)
    if run_time > spawned_process_max_run_time:
        terminate_spawned_process()
    else:
        spawned_process_status_update()


def spawned_process_ended():
    global spawned_process_stop_time
    spawned_process_stop_time = timeit.default_timer()
    run_time = (spawned_process_stop_time - spawned_process_start_time)
    etld_lib_functions.logger.info(f"Final Runtime: {run_time:,.0f} seconds")
    etld_lib_functions.logger.info(f"stop")


def spawned_process_status_update():
    global spawned_process_status_count
    time.sleep(spawned_process_sleep_time)
    if spawned_process_status_count > spawned_process_count_to_status_update or spawned_process_status_count == 0:
        etld_lib_functions.logger.info(
            f"Job PID {str(spawned_process.pid)} {target_module_to_run.__name__} "
            f"job running in background.")
        etld_lib_config.log_system_stats_cpu_ram_swap_disk()
        spawned_process_status_count = 1
        if etl_files_are_stalled():
            terminate_spawned_process_due_to_stalled_output()
    else:
        spawned_process_status_count = spawned_process_status_count + 1


def terminate_spawned_process_due_to_stalled_output():
    etld_lib_functions.logger.error(f"Stalled Error: {etl_files_max_time_between_changes:,.0f} Seconds Exceeded")
    etld_lib_functions.logger.error(f"Directory:     {etl_dir_to_monitor} Directory monitored for changes.")
    etld_lib_functions.logger.error(f"Please review previous extract function name in log to identify the "
                                    f"stalling program.")
    etld_lib_functions.logger.error(f"Use api options from the logs to create curl command to simulate issue.")
    etld_lib_functions.logger.error(f"Provide feedback to your Qualys TAM to assist with engaging support "
                                    f"if problem persists.")
    etld_lib_functions.logger.error(f"Terminating job.")
    spawned_process.terminate()
    spawned_process.join()
    exit(1)


def terminate_spawned_process():
    etld_lib_functions.logger.error(f"Max Run Time: {spawned_process_max_run_time:,.0f} Seconds Exceeded")
    etld_lib_functions.logger.error(f"Please review for issues that are slowing down the program.")
    etld_lib_functions.logger.error(f"Terminating job.")
    spawned_process.terminate()
    spawned_process.join()
    exit(1)


def rotate_log_check():
    if log_file_path.is_file():
        log_file_size = log_file_path.stat().st_size  # In Bytes
        if log_file_size > log_file_max_size:
            shutil.copy2(log_file_path, log_file_rotate_path, follow_symlinks=True)
            fo = open(log_file_path, 'w+')
            fo.close()


def etl_main():
    exception_message = "ERROR: Program already running. Please retry later: "
    try:
        with open(lock_file, 'wb+') as lock_program_fcntl:  # If locked, exit.
            fcntl.flock(lock_program_fcntl, fcntl.LOCK_EX | fcntl.LOCK_NB)
            exception_message = "ERROR: rotate_log_check. Please retry later: "
            rotate_log_check()
            if log_dir.is_dir():
                exception_message = "ERROR: spawn_etl_in_background. Please retry later: "
                with open(log_file_path, 'a', newline='', encoding='utf-8') as log_fo:
                    with redirect_stdout(log_fo), redirect_stderr(sys.stdout):
                        #etld_lib_functions.main(log_level=logging.INFO,
                        #                        my_logger_prog_name=target_module_to_run.__name__)
                        etld_lib_functions.main(my_logger_prog_name=target_module_to_run.__name__)
                        etld_lib_config.main()
                        etld_lib_credentials.main()
                        etl_workflow = target_method_in_module_to_run
                        etld_lib_authentication_objects.main(http_user_agent_etl_workflow_message=etl_workflow,test_basic_authorization=True)
                        spawn_etl_in_background()
                        log_fo.flush()
                        time.sleep(1)
                        passed_test = \
                            etld_lib_log_validation.validate_log_has_no_errors(target_method_in_module_to_run)
                        if passed_test:
                            log_fo.flush()
                            etld_lib_functions.logger.info("SUCCESS: validate_log_has_no_errors=True")
                        else:
                            log_fo.flush()
                            exception_message = "ERROR: validate_log_has_no_errors=False"
                            etld_lib_functions.logger.error(exception_message)
                            raise Exception(exception_message)
            else:
                exception_message = f"ERROR: logdir_missing: {log_dir} - Potential permissions issue."
                raise Exception(exception_message)

    except Exception as e:
        print(f"{exception_message} {__file__} ")
        print(f"Exception: {e}")
        exit(1)



import os
from pathlib import Path
from qualys_etl.etld_lib import etld_lib_functions
from qualys_etl.etld_lib import etld_lib_config


def display_log_error_messages(error_messages_list):
    for error_message in error_messages_list:
        etld_lib_functions.logger.error(f"{error_message}")


def get_workflow_log_entry_dict(log_entry_line):
    log_entry_row_list = str(log_entry_line).split('|')
    log_entry_dict = {}
    if len(log_entry_row_list) == len(etld_lib_config.run_log_csv_columns()) and \
            log_entry_row_list[0].startswith("20"):
        for csv_index in range(len(etld_lib_config.run_log_csv_columns())):
            column_name = etld_lib_config.run_log_csv_columns()[csv_index]
            log_entry_dict[f'{column_name}'] = str(log_entry_row_list[csv_index]).strip()

        log_workflow_list =  str(log_entry_dict['LOG_WORKFLOW']).split(':')
        if len(log_workflow_list) != 2:
            log_entry_dict['LOG_WORKFLOW_NAME'] = log_workflow_list[0]
            log_entry_dict['LOG_WORKFLOW_DATE'] = log_workflow_list[0]
        else:
            log_entry_dict['LOG_WORKFLOW_NAME'] = log_workflow_list[0]
            log_entry_dict['LOG_WORKFLOW_DATE'] = log_workflow_list[1]

        log_entry_dict['LOG_ROW_STRUCTURE_ISVALID'] = True
    else:
        log_entry_string = "".join(log_entry_line).strip()
        log_entry_dict['LOG_MESSAGE'] = f"UNBOUND LOG ERROR: {log_entry_string}"
        log_entry_dict['LOG_LEVEL'] = "ERROR"
        log_entry_dict['LOG_ROW_STRUCTURE_ISVALID'] = False

    return log_entry_dict


def get_workflow_log_last_entry(log_file):
    if not Path(log_file).is_file():
        etld_lib_functions.logger.error(f"Log File Doesn't Exist: {str(log_file)}, rerun after executing corresponding etl")
        raise FileNotFoundError(f"{str(log_file)}, rerun after executing corresponding etl")

    with open(str(log_file), "rb") as log_file_handle:
        try:
            log_file_handle.seek(-2, os.SEEK_END)
            while log_file_handle.read(1) != b'\n':
                log_file_handle.seek(-2, os.SEEK_CUR)
        except OSError:
            log_file_handle.seek(0)
        log_entry = log_file_handle.readline().decode()
        log_entry_dict = get_workflow_log_entry_dict(log_entry)
    return log_entry_dict


def get_workflow_log(log_file_path):
    last_log_entry_dict = get_workflow_log_last_entry(log_file_path)
    all_log_entries = []
    found_first_log_workflow = False
    if last_log_entry_dict['LOG_ROW_STRUCTURE_ISVALID']:
        with open(str(log_file_path), "rt", encoding='utf-8') as read_file:
            for log_entry_line in read_file:
                log_entry_dict = get_workflow_log_entry_dict(log_entry_line)
                if log_entry_dict['LOG_ROW_STRUCTURE_ISVALID']:
                    if str(log_entry_dict['LOG_WORKFLOW']).strip() == str(last_log_entry_dict['LOG_WORKFLOW']).strip():
                        all_log_entries.append(log_entry_dict)
                        found_first_log_workflow = True
                elif found_first_log_workflow:
                    #log_entry_dict = craft_log_out_of_bound_entry(last_log_entry_dict)
                    # Keep 'LOG_MESSAGE', 'LOG_LEVEL', change ['LOG_ROW_STRUCTURE_ISVALID'] = True
                    log_entry_dict['LOG_DATETIME'] = last_log_entry_dict['LOG_DATETIME']
                    log_entry_dict['LOG_WORKFLOW'] = last_log_entry_dict['LOG_WORKFLOW']
                    log_entry_dict['LOG_WORKFLOW_NAME'] = last_log_entry_dict['LOG_WORKFLOW_NAME']
                    log_entry_dict['LOG_WORKFLOW_DATE'] = last_log_entry_dict['LOG_WORKFLOW_DATE']
                    log_entry_dict['LOG_USERNAME'] = last_log_entry_dict['LOG_USERNAME']
                    log_entry_dict['LOG_FUNCTION'] = 'etld_lib_log_validation_out_of_bounds'
                    log_entry_dict['LOG_ROW_STRUCTURE_ISVALID'] = True
                    all_log_entries.append(log_entry_dict)
    else:
        #log_entry_dict = craft_log_out_of_bound_entry(last_log_entry_dict)
        # Completely out of bounds error.
        log_entry_dict = {
            'LOG_DATETIME': 'UNKNOWN',
            'LOG_LEVEL': 'ERROR',
            'LOG_MESSAGE': 'LAST LINE OF LOG FILE IS IN ERROR, PLEASE INVESTIGATE ISSUE',
            'LOG_WORKFLOW': 'UNKNOWN:UNKNOWN',
            'LOG_WORKFLOW_NAME': 'UNKNOWN',
            'LOG_WORKFLOW_DATE': 'UNKNOWN',
            'LOG_USERNAME': 'UNKNOWN',
            'LOG_FUNCTION': 'etld_lib_log_validation_completely_out_of_bounds',
            'LOG_ROW_STRUCTURE_ISVALID': True
        }
        all_log_entries.append(log_entry_dict)
    return all_log_entries


def test_workflow_spawned_process_ended(all_log_entries: list):
    etl_workflow_completed = False
    for log_entry_dict in all_log_entries:
        if str(log_entry_dict['LOG_FUNCTION']).strip() == 'spawned_process_ended':
            etl_workflow_completed = True

    if etl_workflow_completed:
        etld_lib_functions.logger.info(f"Job completed, spawned_process_ended function found.")
    else:
        etld_lib_functions.logger.error(f"ERROR: Job did not complete, spawned_process_ended function not found.")

    return etl_workflow_completed


def test_workflow_log_no_errors_found_thus_far(all_log_entries: list):
    etl_workflow_valid_flag = True
    error_messages_list = []
    for log_entry_dict in all_log_entries:
        if log_entry_dict['LOG_LEVEL'] == 'ERROR':
            error_messages_list.append(f"ERROR: {log_entry_dict['LOG_MESSAGE']}")
            etl_workflow_valid_flag = False

    if etl_workflow_valid_flag:
        etld_lib_functions.logger.info(f"Passed Tests: {all_log_entries[0]['LOG_WORKFLOW']} ")
    else:
        display_log_error_messages(error_messages_list=error_messages_list)

    return etl_workflow_valid_flag


def test_workflow_log_completed_successfully(all_log_entries: list):
    etl_workflow_valid_flag = True
    error_messages_list = []
    for log_entry_dict in all_log_entries:
        if log_entry_dict['LOG_LEVEL'] == 'ERROR':
            error_messages_list.append(f"ERROR: {log_entry_dict['LOG_MESSAGE']}")
            etl_workflow_valid_flag = False

    if etl_workflow_valid_flag:
        etld_lib_functions.logger.info(f"Passed Tests: {all_log_entries[0]['LOG_WORKFLOW']} ")
    else:
        display_log_error_messages(error_messages_list=error_messages_list)

    if test_workflow_spawned_process_ended(all_log_entries):
        pass
    else:
        etl_workflow_valid_flag = False

    return etl_workflow_valid_flag


def validate_log_has_no_errors(etl_workflow_option):
    etl_workflow_location_dict = \
        etld_lib_config.get_etl_workflow_data_location_dict(etl_workflow_option)
    all_log_entries = get_workflow_log(log_file_path=Path(etl_workflow_location_dict['log_file']))
    return test_workflow_log_completed_successfully(all_log_entries)


def validate_log_has_no_errors_prior_to_distribution(etl_workflow_option):
    # Called by in process programs before end of job.
    etl_workflow_location_dict = \
        etld_lib_config.get_etl_workflow_data_location_dict(etl_workflow_option)
    all_log_entries = get_workflow_log(log_file_path=Path(etl_workflow_location_dict['log_file']))
    return test_workflow_log_no_errors_found_thus_far(all_log_entries)


def main_validate_log_has_no_errors(etl_workflow):
    etld_lib_functions.main(my_logger_prog_name='main_validate_log_has_no_errors')
    etld_lib_config.main()
    passed_all_tests = validate_log_has_no_errors(etl_workflow_option=etl_workflow)
    if passed_all_tests:
       pass
    else:
       exit(2)

    return passed_all_tests


def main_validate_all_logs_have_no_errors():
    etld_lib_functions.main(my_logger_prog_name='main_validate_all_logs_have_no_errors')
    etld_lib_config.main()
    main()


def main():
    final_exit_status = 0
    etl_workflow_list = etld_lib_config.etl_workflow_list
    etl_workflow_list = ['test_system_etl_workflow']
    for etl_workflow in etl_workflow_list:
        passed_all_tests = \
            validate_log_has_no_errors(etl_workflow_option=etl_workflow)
        if passed_all_tests:
            pass
        else:
            final_exit_status = 2
    exit(final_exit_status)


if __name__ == "__main__":
    etld_lib_functions.main(my_logger_prog_name='etld_lib_log_validation')
    etld_lib_config.main()
    main()
